package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Annotation;
import org.eclipse.jdt.internal.compiler.util.SimpleLookupTable;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.ReferenceExpression;
import org.eclipse.jdt.internal.compiler.ast.ExpressionContext;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.CaseStatement;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.util.HashtableOfObject;
import org.eclipse.jdt.internal.compiler.ast.ImportReference;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import org.eclipse.jdt.internal.compiler.ast.AbstractVariableDeclaration;
import org.eclipse.jdt.internal.compiler.ast.NameReference;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import org.eclipse.jdt.internal.compiler.util.ObjectVector;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.ast.LambdaExpression;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import java.util.Map;
import org.eclipse.jdt.internal.compiler.util.SimpleSet;
import java.util.HashMap;
import org.eclipse.jdt.internal.compiler.ast.TypeParameter;
import org.eclipse.jdt.internal.compiler.ast.Invocation;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import java.util.HashSet;
import java.util.Set;

public abstract class Scope
{
    public static final int BLOCK_SCOPE = 1;
    public static final int CLASS_SCOPE = 3;
    public static final int COMPILATION_UNIT_SCOPE = 4;
    public static final int METHOD_SCOPE = 2;
    public static final int NOT_COMPATIBLE = -1;
    public static final int COMPATIBLE = 0;
    public static final int AUTOBOX_COMPATIBLE = 1;
    public static final int VARARGS_COMPATIBLE = 2;
    public static final int EQUAL_OR_MORE_SPECIFIC = -1;
    public static final int NOT_RELATED = 0;
    public static final int MORE_GENERIC = 1;
    public int kind;
    public Scope parent;
    private static Scope.Substitutor defaultSubstitutor;
    
    static {
        Scope.defaultSubstitutor = new Scope.Substitutor();
    }
    
    protected Scope(int kind, Scope parent) {
        this.kind = kind;
        this.parent = parent;
    }
    
    public static int compareTypes(TypeBinding left, TypeBinding right) {
        if (left.isCompatibleWith(right)) {
            return -1;
        }
        if (right.isCompatibleWith(left)) {
            return 1;
        }
        return 0;
    }
    
    public static TypeBinding convertEliminatingTypeVariables(TypeBinding originalType, ReferenceBinding genericType, int rank, Set eliminatedVariables) {
        if ((originalType.tagBits & 0x20000000L) != 0x0L) {
            switch (originalType.kind()) {
                case 68: {
                    ArrayBinding originalArrayType = (ArrayBinding)originalType;
                    TypeBinding originalLeafComponentType = originalArrayType.leafComponentType;
                    TypeBinding substitute = convertEliminatingTypeVariables(originalLeafComponentType, genericType, rank, eliminatedVariables);
                    if (TypeBinding.notEquals(substitute, originalLeafComponentType)) {
                        return originalArrayType.environment.createArrayType(substitute.leafComponentType(), substitute.dimensions() + originalArrayType.dimensions());
                    }
                    break;
                }
                case 260: {
                    ParameterizedTypeBinding paramType = (ParameterizedTypeBinding)originalType;
                    ReferenceBinding substitutedEnclosing;
                    ReferenceBinding originalEnclosing = substitutedEnclosing = paramType.enclosingType();
                    if (originalEnclosing != null) {
                        substitutedEnclosing = (ReferenceBinding)convertEliminatingTypeVariables(originalEnclosing, genericType, rank, eliminatedVariables);
                    }
                    TypeBinding[] substitutedArguments;
                    TypeBinding[] originalArguments = substitutedArguments = paramType.arguments;
                    for (int i = 0, length = (originalArguments == null) ? 0 : originalArguments.length; i < length; ++i) {
                        TypeBinding originalArgument = originalArguments[i];
                        TypeBinding substitutedArgument = convertEliminatingTypeVariables(originalArgument, paramType.genericType(), i, eliminatedVariables);
                        if (TypeBinding.notEquals(substitutedArgument, originalArgument)) {
                            if (substitutedArguments == originalArguments) {
                                System.arraycopy(originalArguments, 0, substitutedArguments = new TypeBinding[length], 0, i);
                            }
                            substitutedArguments[i] = substitutedArgument;
                        }
                        else if (substitutedArguments != originalArguments) {
                            substitutedArguments[i] = originalArgument;
                        }
                    }
                    if (TypeBinding.notEquals(originalEnclosing, substitutedEnclosing) || originalArguments != substitutedArguments) {
                        return paramType.environment.createParameterizedType(paramType.genericType(), substitutedArguments, substitutedEnclosing);
                    }
                    break;
                }
                case 4100: {
                    if (genericType != null) {
                        TypeVariableBinding originalVariable = (TypeVariableBinding)originalType;
                        if (eliminatedVariables != null && eliminatedVariables.contains(originalType)) {
                            return originalVariable.environment.createWildcard(genericType, rank, null, null, 0);
                        }
                        TypeBinding originalUpperBound = originalVariable.upperBound();
                        if (eliminatedVariables == null) {
                            eliminatedVariables = new HashSet(2);
                        }
                        eliminatedVariables.add(originalVariable);
                        TypeBinding substitutedUpperBound = convertEliminatingTypeVariables(originalUpperBound, genericType, rank, eliminatedVariables);
                        eliminatedVariables.remove(originalVariable);
                        return originalVariable.environment.createWildcard(genericType, rank, substitutedUpperBound, null, 1);
                    }
                }
                case 1028: {
                    break;
                }
                case 2052: {
                    ReferenceBinding currentType = (ReferenceBinding)originalType;
                    ReferenceBinding substitutedEnclosing;
                    ReferenceBinding originalEnclosing = substitutedEnclosing = currentType.enclosingType();
                    if (originalEnclosing != null) {
                        substitutedEnclosing = (ReferenceBinding)convertEliminatingTypeVariables(originalEnclosing, genericType, rank, eliminatedVariables);
                    }
                    TypeBinding[] substitutedArguments;
                    TypeBinding[] originalArguments = substitutedArguments = currentType.typeVariables();
                    for (int j = 0, length2 = (originalArguments == null) ? 0 : originalArguments.length; j < length2; ++j) {
                        TypeBinding originalArgument2 = originalArguments[j];
                        TypeBinding substitutedArgument2 = convertEliminatingTypeVariables(originalArgument2, currentType, j, eliminatedVariables);
                        if (TypeBinding.notEquals(substitutedArgument2, originalArgument2)) {
                            if (substitutedArguments == originalArguments) {
                                System.arraycopy(originalArguments, 0, substitutedArguments = new TypeBinding[length2], 0, j);
                            }
                            substitutedArguments[j] = substitutedArgument2;
                        }
                        else if (substitutedArguments != originalArguments) {
                            substitutedArguments[j] = originalArgument2;
                        }
                    }
                    if (TypeBinding.notEquals(originalEnclosing, substitutedEnclosing) || originalArguments != substitutedArguments) {
                        return ((TypeVariableBinding)originalArguments[0]).environment.createParameterizedType(genericType, substitutedArguments, substitutedEnclosing);
                    }
                    break;
                }
                case 516: {
                    WildcardBinding wildcard = (WildcardBinding)originalType;
                    TypeBinding substitutedBound;
                    TypeBinding originalBound = substitutedBound = wildcard.bound;
                    if (originalBound != null) {
                        substitutedBound = convertEliminatingTypeVariables(originalBound, genericType, rank, eliminatedVariables);
                        if (TypeBinding.notEquals(substitutedBound, originalBound)) {
                            return wildcard.environment.createWildcard(wildcard.genericType, wildcard.rank, substitutedBound, null, wildcard.boundKind);
                        }
                        break;
                    }
                    else {
                        break;
                    }
                }
                case 8196: {
                    WildcardBinding intersection = (WildcardBinding)originalType;
                    TypeBinding substitutedBound;
                    TypeBinding originalBound = substitutedBound = intersection.bound;
                    if (originalBound != null) {
                        substitutedBound = convertEliminatingTypeVariables(originalBound, genericType, rank, eliminatedVariables);
                    }
                    TypeBinding[] substitutedOtherBounds;
                    TypeBinding[] originalOtherBounds = substitutedOtherBounds = intersection.otherBounds;
                    for (int k = 0, length3 = (originalOtherBounds == null) ? 0 : originalOtherBounds.length; k < length3; ++k) {
                        TypeBinding originalOtherBound = originalOtherBounds[k];
                        TypeBinding substitutedOtherBound = convertEliminatingTypeVariables(originalOtherBound, genericType, rank, eliminatedVariables);
                        if (TypeBinding.notEquals(substitutedOtherBound, originalOtherBound)) {
                            if (substitutedOtherBounds == originalOtherBounds) {
                                System.arraycopy(originalOtherBounds, 0, substitutedOtherBounds = new TypeBinding[length3], 0, k);
                            }
                            substitutedOtherBounds[k] = substitutedOtherBound;
                        }
                        else if (substitutedOtherBounds != originalOtherBounds) {
                            substitutedOtherBounds[k] = originalOtherBound;
                        }
                    }
                    if (TypeBinding.notEquals(substitutedBound, originalBound) || substitutedOtherBounds != originalOtherBounds) {
                        return intersection.environment.createWildcard(intersection.genericType, intersection.rank, substitutedBound, substitutedOtherBounds, intersection.boundKind);
                    }
                    break;
                }
            }
        }
        return originalType;
    }
    
    public static TypeBinding getBaseType(char[] name) {
        int length = name.length;
        if (length > 2 && length < 8) {
            switch (name[0]) {
                case 'i': {
                    if (length == 3 && name[1] == 'n' && name[2] == 't') {
                        return TypeBinding.INT;
                    }
                    break;
                }
                case 'v': {
                    if (length == 4 && name[1] == 'o' && name[2] == 'i' && name[3] == 'd') {
                        return TypeBinding.VOID;
                    }
                    break;
                }
                case 'b': {
                    if (length == 7 && name[1] == 'o' && name[2] == 'o' && name[3] == 'l' && name[4] == 'e' && name[5] == 'a' && name[6] == 'n') {
                        return TypeBinding.BOOLEAN;
                    }
                    if (length == 4 && name[1] == 'y' && name[2] == 't' && name[3] == 'e') {
                        return TypeBinding.BYTE;
                    }
                    break;
                }
                case 'c': {
                    if (length == 4 && name[1] == 'h' && name[2] == 'a' && name[3] == 'r') {
                        return TypeBinding.CHAR;
                    }
                    break;
                }
                case 'd': {
                    if (length == 6 && name[1] == 'o' && name[2] == 'u' && name[3] == 'b' && name[4] == 'l' && name[5] == 'e') {
                        return TypeBinding.DOUBLE;
                    }
                    break;
                }
                case 'f': {
                    if (length == 5 && name[1] == 'l' && name[2] == 'o' && name[3] == 'a' && name[4] == 't') {
                        return TypeBinding.FLOAT;
                    }
                    break;
                }
                case 'l': {
                    if (length == 4 && name[1] == 'o' && name[2] == 'n' && name[3] == 'g') {
                        return TypeBinding.LONG;
                    }
                    break;
                }
                case 's': {
                    if (length == 5 && name[1] == 'h' && name[2] == 'o' && name[3] == 'r' && name[4] == 't') {
                        return TypeBinding.SHORT;
                    }
                    break;
                }
            }
        }
        return null;
    }
    
    public static ReferenceBinding[] greaterLowerBound(ReferenceBinding[] types) {
        if (types == null) {
            return null;
        }
        int length = types.length;
        if (length == 0) {
            return null;
        }
        ReferenceBinding[] result = types;
        int removed = 0;
        for (int i = 0; i < length; ++i) {
            ReferenceBinding iType = result[i];
            if (iType != null) {
                for (int j = 0; j < length; ++j) {
                    if (i != j) {
                        ReferenceBinding jType = result[j];
                        if (jType != null && iType.isCompatibleWith(jType)) {
                            if (result == types) {
                                System.arraycopy(result, 0, result = new ReferenceBinding[length], 0, length);
                            }
                            result[j] = null;
                            ++removed;
                        }
                    }
                }
            }
        }
        if (removed == 0) {
            return result;
        }
        if (length == removed) {
            return null;
        }
        ReferenceBinding[] trimmedResult = new ReferenceBinding[length - removed];
        int k = 0;
        int index = 0;
        while (k < length) {
            ReferenceBinding iType2 = result[k];
            if (iType2 != null) {
                trimmedResult[index++] = iType2;
            }
            ++k;
        }
        return trimmedResult;
    }
    
    public static TypeBinding[] greaterLowerBound(TypeBinding[] types, Scope scope, LookupEnvironment environment) {
        if (types == null) {
            return null;
        }
        int length = types.length;
        if (length == 0) {
            return null;
        }
        TypeBinding[] result = types;
        int removed = 0;
        for (int i = 0; i < length; ++i) {
            TypeBinding iType = result[i];
            if (iType != null) {
                for (int j = 0; j < length; ++j) {
                    if (i != j) {
                        TypeBinding jType = result[j];
                        if (jType != null) {
                            if (iType.isCompatibleWith(jType, scope)) {
                                if (result == types) {
                                    System.arraycopy(result, 0, result = new TypeBinding[length], 0, length);
                                }
                                result[j] = null;
                                ++removed;
                            }
                            else if (!jType.isCompatibleWith(iType, scope) && iType.isParameterizedType() && jType.isParameterizedType()) {
                                ParameterizedTypeBinding wideType;
                                ParameterizedTypeBinding narrowType;
                                if (iType.original().isCompatibleWith(jType.original(), scope)) {
                                    wideType = (ParameterizedTypeBinding)jType;
                                    narrowType = (ParameterizedTypeBinding)iType;
                                }
                                else if (jType.original().isCompatibleWith(iType.original(), scope)) {
                                    wideType = (ParameterizedTypeBinding)iType;
                                    narrowType = (ParameterizedTypeBinding)jType;
                                }
                                else {
                                    continue;
                                }
                                if (wideType.arguments != null && narrowType.isProperType(false) && wideType.isProperType(false)) {
                                    int numTypeArgs = wideType.arguments.length;
                                    TypeBinding[] bounds = new TypeBinding[numTypeArgs];
                                    for (int k = 0; k < numTypeArgs; ++k) {
                                        TypeBinding argument = wideType.arguments[k];
                                        bounds[k] = (argument.isTypeVariable() ? ((TypeVariableBinding)argument).upperBound() : argument);
                                    }
                                    ReferenceBinding wideOriginal = (ReferenceBinding)wideType.original();
                                    TypeBinding substitutedWideType = environment.createParameterizedType(wideOriginal, bounds, wideOriginal.enclosingType());
                                    if (!narrowType.isCompatibleWith(substitutedWideType, scope)) {
                                        return null;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        if (removed == 0) {
            return result;
        }
        if (length == removed) {
            return null;
        }
        TypeBinding[] trimmedResult = new TypeBinding[length - removed];
        int l = 0;
        int index = 0;
        while (l < length) {
            TypeBinding iType2 = result[l];
            if (iType2 != null) {
                trimmedResult[index++] = iType2;
            }
            ++l;
        }
        return trimmedResult;
    }
    
    public static ReferenceBinding[] substitute(Substitution substitution, ReferenceBinding[] originalTypes) {
        return Scope.defaultSubstitutor.substitute(substitution, originalTypes);
    }
    
    public static TypeBinding substitute(Substitution substitution, TypeBinding originalType) {
        return Scope.defaultSubstitutor.substitute(substitution, originalType);
    }
    
    public static TypeBinding[] substitute(Substitution substitution, TypeBinding[] originalTypes) {
        return Scope.defaultSubstitutor.substitute(substitution, originalTypes);
    }
    
    public TypeBinding boxing(TypeBinding type) {
        if (type.isBaseType() || type.kind() == 65540) {
            return this.environment().computeBoxingType(type);
        }
        return type;
    }
    
    public final ClassScope classScope() {
        Scope scope = this;
        while (!(scope instanceof ClassScope)) {
            scope = scope.parent;
            if (scope != null) {
                continue;
            }
            return null;
        }
        return (ClassScope)scope;
    }
    
    public final CompilationUnitScope compilationUnitScope() {
        Scope lastScope = null;
        Scope scope = this;
        do {
            lastScope = scope;
            scope = scope.parent;
        } while (scope != null);
        return (CompilationUnitScope)lastScope;
    }
    
    public boolean isLambdaScope() {
        return false;
    }
    
    public boolean isLambdaSubscope() {
        Scope scope = this;
        while (scope != null) {
            switch (scope.kind) {
                case 1: {
                    scope = scope.parent;
                    continue;
                }
                case 2: {
                    return scope.isLambdaScope();
                }
                default: {
                    return false;
                }
            }
        }
        return false;
    }
    
    public final CompilerOptions compilerOptions() {
        return this.compilationUnitScope().environment.globalOptions;
    }
    
    protected final MethodBinding computeCompatibleMethod(MethodBinding method, TypeBinding[] arguments, InvocationSite invocationSite) {
        return this.computeCompatibleMethod(method, arguments, invocationSite, false);
    }
    
    protected final MethodBinding computeCompatibleMethod(MethodBinding method, TypeBinding[] arguments, InvocationSite invocationSite, boolean tiebreakingVarargsMethods) {
        TypeBinding[] genericTypeArguments = invocationSite.genericTypeArguments();
        TypeBinding[] parameters = method.parameters;
        TypeVariableBinding[] typeVariables = method.typeVariables;
        if (parameters == arguments && (method.returnType.tagBits & 0x20000000L) == 0x0L && genericTypeArguments == null && typeVariables == Binding.NO_TYPE_VARIABLES) {
            return method;
        }
        int argLength = arguments.length;
        int paramLength = parameters.length;
        boolean isVarArgs = method.isVarargs();
        if (argLength != paramLength && (!isVarArgs || argLength < paramLength - 1)) {
            return null;
        }
        CompilerOptions compilerOptions = this.compilerOptions();
        if (typeVariables != Binding.NO_TYPE_VARIABLES && compilerOptions.sourceLevel >= 3211264L) {
            TypeBinding[] newArgs = null;
            if (compilerOptions.sourceLevel < 3407872L || genericTypeArguments != null) {
                for (int i = 0; i < argLength; ++i) {
                    TypeBinding param = (i < paramLength) ? parameters[i] : parameters[paramLength - 1];
                    if (arguments[i].isBaseType() != param.isBaseType()) {
                        if (newArgs == null) {
                            newArgs = new TypeBinding[argLength];
                            System.arraycopy(arguments, 0, newArgs, 0, argLength);
                        }
                        newArgs[i] = this.environment().computeBoxingType(arguments[i]);
                    }
                }
            }
            if (newArgs != null) {
                arguments = newArgs;
            }
            method = ParameterizedGenericMethodBinding.computeCompatibleMethod(method, arguments, this, invocationSite);
            if (method == null) {
                return null;
            }
            if (!method.isValidBinding()) {
                return method;
            }
            if (compilerOptions.sourceLevel >= 3407872L && method instanceof ParameterizedGenericMethodBinding && invocationSite instanceof Invocation) {
                Invocation invocation2 = (Invocation)invocationSite;
                InferenceContext18 infCtx2 = invocation2.getInferenceContext((ParameterizedGenericMethodBinding)method);
                if (infCtx2 != null) {
                    return method;
                }
            }
        }
        else if (genericTypeArguments != null && compilerOptions.complianceLevel < 3342336L) {
            if ((method instanceof ParameterizedGenericMethodBinding)) {
                if (!((ParameterizedGenericMethodBinding)method).wasInferred) {
                    return new ProblemMethodBinding(method, method.selector, genericTypeArguments, 13);
                }
            }
            else if (!method.isOverriding() || !this.isOverriddenMethodGeneric(method)) {
                return new ProblemMethodBinding(method, method.selector, genericTypeArguments, 11);
            }
        }
        else if (typeVariables == Binding.NO_TYPE_VARIABLES && method instanceof ParameterizedGenericMethodBinding && compilerOptions.sourceLevel >= 3407872L && invocationSite instanceof Invocation) {
            Invocation invocation = (Invocation)invocationSite;
            InferenceContext18 infCtx = invocation.getInferenceContext((ParameterizedGenericMethodBinding)method);
            if (infCtx != null) {
                return method;
            }
        }
        if (tiebreakingVarargsMethods && CompilerOptions.tolerateIllegalAmbiguousVarargsInvocation && compilerOptions.complianceLevel < 3342336L) {
            tiebreakingVarargsMethods = false;
        }
        if (this.parameterCompatibilityLevel(method, arguments, tiebreakingVarargsMethods) > -1) {
            if ((method.tagBits & 0x10000000000000L) != 0x0L) {
                return this.environment().createPolymorphicMethod(method, arguments);
            }
            return method;
        }
        else {
            if (genericTypeArguments != null && typeVariables != Binding.NO_TYPE_VARIABLES) {
                return new ProblemMethodBinding(method, method.selector, arguments, 12);
            }
            if ((method instanceof PolyParameterizedGenericMethodBinding)) {
                return new ProblemMethodBinding(method, method.selector, method.parameters, 27);
            }
            return null;
        }
    }
    
    protected boolean connectTypeVariables(TypeParameter[] typeParameters, boolean checkForErasedCandidateCollisions) {
        if (typeParameters == null || typeParameters.length == 0) {
            return true;
        }
        Map<Object, Object> invocations = new HashMap(2);
        boolean noProblems = true;
        int paramLength = typeParameters.length;
        int i = 0;
        while (i < paramLength) {
            TypeParameter typeParameter = typeParameters[i];
            TypeVariableBinding typeVariable = typeParameter.binding;
            if (typeVariable == null) {
                return false;
            }
            typeVariable.setSuperClass(this.getJavaLangObject());
            typeVariable.setSuperInterfaces(Binding.NO_SUPERINTERFACES);
            typeVariable.setFirstBound(null);
            ++i;
        }
        for (i = 0; i < paramLength; ++i) {
            TypeParameter typeParameter = typeParameters[i];
            TypeVariableBinding typeVariable = typeParameter.binding;
            TypeReference typeRef = typeParameter.type;
            if (typeRef != null) {
                boolean isFirstBoundTypeVariable = false;
                TypeBinding superType = (this.kind == 2) ? typeRef.resolveType((BlockScope)this, false, 256) : typeRef.resolveType((ClassScope)this, 256);
                Label_0552: {
                    if (superType == null) {
                        typeVariable.tagBits |= 0x20000L;
                    }
                    else {
                        typeRef.resolvedType = superType;
                        switch (superType.kind()) {
                            case 68: {
                                this.problemReporter().boundCannotBeArray(typeRef, superType);
                                typeVariable.tagBits |= 0x20000L;
                                break Label_0552;
                            }
                            case 4100: {
                                isFirstBoundTypeVariable = true;
                                TypeVariableBinding varSuperType = (TypeVariableBinding)superType;
                                if (varSuperType.rank >= typeVariable.rank && varSuperType.declaringElement == typeVariable.declaringElement && this.compilerOptions().complianceLevel <= 3276800L) {
                                    this.problemReporter().forwardTypeVariableReference(typeParameter, varSuperType);
                                    typeVariable.tagBits |= 0x20000L;
                                    break Label_0552;
                                }
                                if (this.compilerOptions().complianceLevel > 3276800L && typeVariable.rank >= varSuperType.rank && varSuperType.declaringElement == typeVariable.declaringElement) {
                                    SimpleSet set = new SimpleSet(typeParameters.length);
                                    set.add(typeVariable);
                                    ReferenceBinding superBinding = varSuperType;
                                    while (superBinding instanceof TypeVariableBinding) {
                                        if (set.includes(superBinding)) {
                                            this.problemReporter().hierarchyCircularity(typeVariable, varSuperType, typeRef);
                                            typeVariable.tagBits |= 0x20000L;
                                            break Label_0552;
                                        }
                                        set.add(superBinding);
                                        superBinding = ((TypeVariableBinding)superBinding).superclass;
                                    }
                                    break;
                                }
                                break;
                            }
                            default: {
                                if (((ReferenceBinding)superType).isFinal()) {
                                    this.problemReporter().finalVariableBound(typeVariable, typeRef);
                                    break;
                                }
                                break;
                            }
                        }
                        ReferenceBinding superRefType = (ReferenceBinding)superType;
                        if (!superType.isInterface()) {
                            typeVariable.setSuperClass(superRefType);
                        }
                        else {
                            typeVariable.setSuperInterfaces(new ReferenceBinding[] { superRefType });
                        }
                        typeVariable.tagBits |= (superType.tagBits & 0x800L);
                        typeVariable.setFirstBound(superRefType);
                    }
                }
                TypeReference[] boundRefs = typeParameter.bounds;
                if (boundRefs != null) {
                Label_0970:
                    for (int j = 0, boundLength = boundRefs.length; j < boundLength; ++j) {
                        typeRef = boundRefs[j];
                        superType = ((this.kind == 2) ? typeRef.resolveType((BlockScope)this, false) : typeRef.resolveType((ClassScope)this));
                        if (superType == null) {
                            typeVariable.tagBits |= 0x20000L;
                        }
                        else {
                            typeVariable.tagBits |= (superType.tagBits & 0x800L);
                            boolean didAlreadyComplain = !typeRef.resolvedType.isValidBinding();
                            if (isFirstBoundTypeVariable && j == 0) {
                                this.problemReporter().noAdditionalBoundAfterTypeVariable(typeRef);
                                typeVariable.tagBits |= 0x20000L;
                                didAlreadyComplain = true;
                            }
                            else if (superType.isArrayType()) {
                                if (!didAlreadyComplain) {
                                    this.problemReporter().boundCannotBeArray(typeRef, superType);
                                    typeVariable.tagBits |= 0x20000L;
                                    continue;
                                }
                                continue;
                            }
                            else if (!superType.isInterface()) {
                                if (!didAlreadyComplain) {
                                    this.problemReporter().boundMustBeAnInterface(typeRef, superType);
                                    typeVariable.tagBits |= 0x20000L;
                                    continue;
                                }
                                continue;
                            }
                            if (!checkForErasedCandidateCollisions || !TypeBinding.equalsEquals(typeVariable.firstBound, typeVariable.superclass) || !this.hasErasedCandidatesCollisions(superType, typeVariable.superclass, invocations, typeVariable, typeRef)) {
                                ReferenceBinding superRefType2 = (ReferenceBinding)superType;
                                int index = typeVariable.superInterfaces.length;
                                while (--index >= 0) {
                                    ReferenceBinding previousInterface = typeVariable.superInterfaces[index];
                                    if (TypeBinding.equalsEquals(previousInterface, superRefType2)) {
                                        this.problemReporter().duplicateBounds(typeRef, superType);
                                        typeVariable.tagBits |= 0x20000L;
                                        continue Label_0970;
                                    }
                                    if (checkForErasedCandidateCollisions && this.hasErasedCandidatesCollisions(superType, previousInterface, invocations, typeVariable, typeRef)) {
                                        continue Label_0970;
                                    }
                                    continue;
                                }
                                int size = typeVariable.superInterfaces.length;
                                System.arraycopy(typeVariable.superInterfaces, 0, typeVariable.setSuperInterfaces(new ReferenceBinding[size + 1]), 0, size);
                                typeVariable.superInterfaces[size] = superRefType2;
                            }
                        }
                    }
                }
                noProblems &= ((typeVariable.tagBits & 0x20000L) == 0x0L);
            }
        }
        boolean declaresNullTypeAnnotation = false;
        for (int k = 0; k < paramLength; declaresNullTypeAnnotation |= typeParameters[k].binding.hasNullTypeAnnotations(), ++k) {
            this.resolveTypeParameter(typeParameters[k]);
        }
        if (declaresNullTypeAnnotation) {
            for (int k = 0; k < paramLength; ++k) {
                typeParameters[k].binding.updateTagBits();
            }
        }
        return noProblems;
    }
    
    public ArrayBinding createArrayType(TypeBinding type, int dimension) {
        return this.createArrayType(type, dimension, Binding.NO_ANNOTATIONS);
    }
    
    public ArrayBinding createArrayType(TypeBinding type, int dimension, AnnotationBinding[] annotations) {
        if (type.isValidBinding()) {
            return this.environment().createArrayType(type, dimension, annotations);
        }
        return new ArrayBinding(type, dimension, this.environment());
    }
    
    public TypeVariableBinding[] createTypeVariables(TypeParameter[] typeParameters, Binding declaringElement) {
        if (typeParameters == null || typeParameters.length == 0) {
            return Binding.NO_TYPE_VARIABLES;
        }
        PackageBinding unitPackage = this.compilationUnitScope().fPackage;
        int length = typeParameters.length;
        TypeVariableBinding[] typeVariableBindings = new TypeVariableBinding[length];
        int count = 0;
        TypeVariableBinding parameterBinding;
        for (int i = 0; i < length; typeVariableBindings[count++] = parameterBinding, ++i) {
            TypeParameter typeParameter = typeParameters[i];
            parameterBinding = new TypeVariableBinding(typeParameter.name, declaringElement, i, this.environment());
            parameterBinding.fPackage = unitPackage;
            typeParameter.binding = parameterBinding;
            if ((typeParameter.bits & 0x100000) != 0x0) {
                switch (declaringElement.kind()) {
                    case 8: {
                        MethodBinding methodBinding = (MethodBinding)declaringElement;
                        AbstractMethodDeclaration sourceMethod = methodBinding.sourceMethod();
                        if (sourceMethod != null) {
                            sourceMethod.bits |= 0x100000;
                            break;
                        }
                        break;
                    }
                    case 4: {
                        if ((declaringElement instanceof SourceTypeBinding)) {
                            SourceTypeBinding sourceTypeBinding = (SourceTypeBinding)declaringElement;
                            TypeDeclaration typeDeclaration = sourceTypeBinding.scope.referenceContext;
                            if (typeDeclaration != null) {
                                typeDeclaration.bits |= 0x100000;
                                break;
                            }
                            break;
                        }
                        else {
                            break;
                        }
                    }
                }
            }
            for (TypeVariableBinding knownVar : typeVariableBindings) {
                if (CharOperation.equals(knownVar.sourceName, typeParameter.name)) {
                    this.problemReporter().duplicateTypeParameterInType(typeParameter);
                }
            }
        }
        if (count != length) {
            System.arraycopy(typeVariableBindings, 0, typeVariableBindings = new TypeVariableBinding[count], 0, count);
            return typeVariableBindings;
        }
        return typeVariableBindings;
    }
    
    void resolveTypeParameter(TypeParameter typeParameter) {
    }
    
    public final ClassScope enclosingClassScope() {
        Scope scope = this;
        while ((scope = scope.parent) != null) {
            if ((scope instanceof ClassScope)) {
                return (ClassScope)scope;
            }
            continue;
        }
        return null;
    }
    
    public final MethodScope enclosingMethodScope() {
        Scope scope = this;
        while ((scope = scope.parent) != null) {
            if ((scope instanceof MethodScope)) {
                return (MethodScope)scope;
            }
            continue;
        }
        return null;
    }
    
    public final MethodScope enclosingLambdaScope() {
        Scope scope = this;
        while ((scope = scope.parent) != null) {
            if ((scope instanceof MethodScope)) {
                MethodScope methodScope = (MethodScope)scope;
                if ((methodScope.referenceContext instanceof LambdaExpression)) {
                    return methodScope;
                }
                continue;
            }
            else {
                continue;
            }
        }
        return null;
    }
    
    public final ReferenceBinding enclosingReceiverType() {
        Scope scope = this;
        while (!(scope instanceof ClassScope)) {
            scope = scope.parent;
            if (scope != null) {
                continue;
            }
            return null;
        }
        return this.environment().convertToParameterizedType(((ClassScope)scope).referenceContext.binding);
    }
    
    public ReferenceContext enclosingReferenceContext() {
        Scope current = this;
        while ((current = current.parent) != null) {
            switch (current.kind) {
                case 2: {
                    return ((MethodScope)current).referenceContext;
                }
                case 3: {
                    return ((ClassScope)current).referenceContext;
                }
                case 4: {
                    return ((CompilationUnitScope)current).referenceContext;
                }
                default: {
                    continue;
                }
            }
        }
        return null;
    }
    
    public final SourceTypeBinding enclosingSourceType() {
        Scope scope = this;
        while (!(scope instanceof ClassScope)) {
            scope = scope.parent;
            if (scope != null) {
                continue;
            }
            return null;
        }
        return ((ClassScope)scope).referenceContext.binding;
    }
    
    public final LookupEnvironment environment() {
        Scope unitScope;
        Scope scope;
        for (unitScope = this; (scope = unitScope.parent) != null; unitScope = scope) {}
        return ((CompilationUnitScope)unitScope).environment;
    }
    
    protected MethodBinding findDefaultAbstractMethod(ReferenceBinding receiverType, char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite, ReferenceBinding classHierarchyStart, ObjectVector found, MethodBinding[] concreteMatches) {
        int startFoundSize = found.size;
        boolean sourceLevel18 = this.compilerOptions().sourceLevel >= 3407872L;
        ReferenceBinding currentType = classHierarchyStart;
        List<TypeBinding> visitedTypes = new ArrayList();
        while (currentType != null) {
            this.findMethodInSuperInterfaces(currentType, selector, found, visitedTypes, invocationSite);
            currentType = currentType.superclass();
        }
        int candidatesCount = (concreteMatches == null) ? 0 : concreteMatches.length;
        int foundSize = found.size;
        MethodBinding[] candidates = new MethodBinding[foundSize - startFoundSize + candidatesCount];
        if (concreteMatches != null) {
            System.arraycopy(concreteMatches, 0, candidates, 0, candidatesCount);
        }
        MethodBinding problemMethod = null;
        if (foundSize > startFoundSize) {
            MethodVerifier methodVerifier = this.environment().methodVerifier();
        Label_0313:
            for (int i = startFoundSize; i < foundSize; ++i) {
                MethodBinding methodBinding = (MethodBinding)found.elementAt(i);
                MethodBinding compatibleMethod = this.computeCompatibleMethod(methodBinding, argumentTypes, invocationSite);
                if (compatibleMethod != null) {
                    if (compatibleMethod.isValidBinding()) {
                        if (concreteMatches != null) {
                            for (int j = 0, length = concreteMatches.length; j < length; ++j) {
                                methodVerifier.areMethodsCompatible(concreteMatches[j], compatibleMethod);
                            }
                        }
                        if (sourceLevel18 || !compatibleMethod.isVarargs() || !(compatibleMethod instanceof ParameterizedGenericMethodBinding)) {
                            int j = 0;
                            while (j < startFoundSize) {
                                MethodBinding classMethod = (MethodBinding)found.elementAt(j);
                                if (classMethod != null && methodVerifier.areMethodsCompatible(classMethod, compatibleMethod)) {
                                    continue Label_0313;
                                }
                                ++j;
                            }
                        }
                        candidates[candidatesCount++] = compatibleMethod;
                    }
                    else if (problemMethod == null) {
                        problemMethod = compatibleMethod;
                    }
                }
            }
        }
        MethodBinding concreteMatch = null;
        if (candidatesCount < 2) {
            if (concreteMatches == null && candidatesCount == 0) {
                return problemMethod;
            }
            concreteMatch = candidates[0];
            if (concreteMatch != null) {
                this.compilationUnitScope().recordTypeReferences(concreteMatch.thrownExceptions);
                return concreteMatch;
            }
            return concreteMatch;
        }
        else {
            if (this.compilerOptions().complianceLevel >= 3145728L) {
                return this.mostSpecificMethodBinding(candidates, candidatesCount, argumentTypes, invocationSite, receiverType);
            }
            return this.mostSpecificInterfaceMethodBinding(candidates, candidatesCount, invocationSite);
        }
    }
    
    public ReferenceBinding findDirectMemberType(char[] typeName, ReferenceBinding enclosingType) {
        if ((enclosingType.tagBits & 0x10000L) != 0x0L) {
            return null;
        }
        ReferenceBinding enclosingReceiverType = this.enclosingReceiverType();
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordReference(enclosingType, typeName);
        ReferenceBinding memberType = enclosingType.getMemberType(typeName);
        if (memberType != null) {
            unitScope.recordTypeReference(memberType);
            if (enclosingReceiverType == null) {
                if (memberType.canBeSeenBy(this.getCurrentPackage())) {
                    return memberType;
                }
                if ((this instanceof CompilationUnitScope)) {
                    TypeDeclaration[] types = ((CompilationUnitScope)this).referenceContext.types;
                    if (types != null) {
                        int i = 0;
                        int max = types.length;
                        while (i < max) {
                            if (memberType.canBeSeenBy(enclosingType, types[i].binding)) {
                                return memberType;
                            }
                            ++i;
                        }
                    }
                }
            }
            else if (memberType.canBeSeenBy(enclosingType, enclosingReceiverType)) {
                return memberType;
            }
            return new ProblemReferenceBinding(new char[][] { typeName }, memberType, 2);
        }
        else {
            return null;
        }
    }
    
    public MethodBinding findExactMethod(ReferenceBinding receiverType, char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite) {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordTypeReferences(argumentTypes);
        MethodBinding exactMethod = receiverType.getExactMethod(selector, argumentTypes, unitScope);
        if (exactMethod != null && exactMethod.typeVariables == Binding.NO_TYPE_VARIABLES && !exactMethod.isBridge()) {
            if (this.compilerOptions().sourceLevel >= 3211264L) {
                int i = argumentTypes.length;
                while (--i >= 0) {
                    TypeBinding t = argumentTypes[i].leafComponentType();
                    if (!(t instanceof ReferenceBinding)) {
                        continue;
                    }
                    ReferenceBinding r = (ReferenceBinding)t;
                    if (r.isHierarchyConnected()) {
                        if (this.isSubtypeOfRawType(r)) {
                            return null;
                        }
                        continue;
                    }
                    else {
                        if (r.isRawType()) {
                            return null;
                        }
                        continue;
                    }
                }
            }
            unitScope.recordTypeReferences(exactMethod.thrownExceptions);
            if (exactMethod.isAbstract() && exactMethod.thrownExceptions != Binding.NO_EXCEPTIONS) {
                return null;
            }
            if (exactMethod.canBeSeenBy(receiverType, invocationSite, this)) {
                if (argumentTypes == Binding.NO_PARAMETERS && CharOperation.equals(selector, TypeConstants.GETCLASS) && exactMethod.returnType.isParameterizedType()) {
                    return this.environment().createGetClassMethod(receiverType, exactMethod, this);
                }
                if (invocationSite.genericTypeArguments() != null) {
                    exactMethod = this.computeCompatibleMethod(exactMethod, argumentTypes, invocationSite);
                }
                else if ((exactMethod.tagBits & 0x10000000000000L) != 0x0L) {
                    return this.environment().createPolymorphicMethod(exactMethod, argumentTypes);
                }
                return exactMethod;
            }
        }
        return null;
    }
    
    public FieldBinding findField(TypeBinding receiverType, char[] fieldName, InvocationSite invocationSite, boolean needResolve) {
        return this.findField(receiverType, fieldName, invocationSite, needResolve, false);
    }
    
    public FieldBinding findField(TypeBinding receiverType, char[] fieldName, InvocationSite invocationSite, boolean needResolve, boolean invisibleFieldsOk) {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordTypeReference(receiverType);
        Label_0189: {
            TypeBinding leafType;
            switch (receiverType.kind()) {
                case 132: {
                    return null;
                }
                case 516:
                case 4100:
                case 8196: {
                    TypeBinding receiverErasure = receiverType.erasure();
                    if (!receiverErasure.isArrayType()) {
                        break Label_0189;
                    }
                    leafType = receiverErasure.leafComponentType();
                    break;
                }
                case 68: {
                    leafType = receiverType.leafComponentType();
                    break;
                }
                default: {
                    break Label_0189;
                }
            }
            if (leafType instanceof ReferenceBinding && !((ReferenceBinding)leafType).canBeSeenBy(this)) {
                return new ProblemFieldBinding((ReferenceBinding)leafType, fieldName, 8);
            }
            if (CharOperation.equals(fieldName, TypeConstants.LENGTH)) {
                if ((leafType.tagBits & 0x80L) != 0x0L) {
                    return new ProblemFieldBinding(ArrayBinding.ArrayLength, null, fieldName, 1);
                }
                return ArrayBinding.ArrayLength;
            }
            else {
                return null;
            }
        }
        ReferenceBinding currentType = (ReferenceBinding)receiverType;
        if (!currentType.canBeSeenBy(this)) {
            return new ProblemFieldBinding(currentType, fieldName, 8);
        }
        currentType.initializeForStaticImports();
        FieldBinding field = currentType.getField(fieldName, needResolve);
        boolean insideTypeAnnotations = this instanceof MethodScope && ((MethodScope)this).insideTypeAnnotation;
        if (field != null) {
            if (invisibleFieldsOk) {
                return field;
            }
            if ((invocationSite != null && !insideTypeAnnotations) ? (!field.canBeSeenBy(currentType, invocationSite, this)) : (!field.canBeSeenBy(this.getCurrentPackage()))) {
                return new ProblemFieldBinding(field, field.declaringClass, fieldName, 2);
            }
            return field;
        }
        else {
            ReferenceBinding[] interfacesToVisit = null;
            int nextPosition = 0;
            FieldBinding visibleField = null;
            boolean keepLooking = true;
            FieldBinding notVisibleField = null;
            while (keepLooking) {
                ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
                if (itsInterfaces != null && itsInterfaces != Binding.NO_SUPERINTERFACES) {
                    if (interfacesToVisit == null) {
                        interfacesToVisit = itsInterfaces;
                        nextPosition = interfacesToVisit.length;
                    }
                    else {
                        int itsLength = itsInterfaces.length;
                        if (nextPosition + itsLength >= interfacesToVisit.length) {
                            System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
                        }
                        int a = 0;
                    Label_0476:
                        while (a < itsLength) {
                            ReferenceBinding next = itsInterfaces[a];
                            int b = 0;
                            Label_0473: {
                                while (b < nextPosition) {
                                    if (TypeBinding.equalsEquals(next, interfacesToVisit[b])) {
                                        ++a;
                                        continue Label_0476;
                                    }
                                    ++b;
                                }
                                interfacesToVisit[nextPosition++] = next;
                                continue;
                            }
                        }
                    }
                }
                if ((currentType = currentType.superclass()) == null) {
                    break;
                }
                unitScope.recordTypeReference(currentType);
                currentType.initializeForStaticImports();
                currentType = (ReferenceBinding)currentType.capture(this, (invocationSite == null) ? 0 : invocationSite.sourceStart(), (invocationSite == null) ? 0 : invocationSite.sourceEnd());
                if ((field = currentType.getField(fieldName, needResolve)) != null) {
                    if (invisibleFieldsOk) {
                        return field;
                    }
                    keepLooking = false;
                    if (field.canBeSeenBy(receiverType, invocationSite, this)) {
                        if (visibleField == null) {
                            visibleField = field;
                        }
                        else {
                            return new ProblemFieldBinding(visibleField, visibleField.declaringClass, fieldName, 3);
                        }
                    }
                    else if (notVisibleField == null) {
                        notVisibleField = field;
                    }
                    else {
                        continue;
                    }
                }
                else {
                    continue;
                }
            }
            if (interfacesToVisit != null) {
                ProblemFieldBinding ambiguous = null;
                for (int i = 0; i < nextPosition; ++i) {
                    ReferenceBinding anInterface = interfacesToVisit[i];
                    unitScope.recordTypeReference(anInterface);
                    if ((field = anInterface.getField(fieldName, true)) != null) {
                        if (invisibleFieldsOk) {
                            return field;
                        }
                        if (visibleField == null) {
                            visibleField = field;
                        }
                        else {
                            ambiguous = new ProblemFieldBinding(visibleField, visibleField.declaringClass, fieldName, 3);
                            break;
                        }
                    }
                    else {
                        ReferenceBinding[] itsInterfaces2 = anInterface.superInterfaces();
                        if (itsInterfaces2 != null && itsInterfaces2 != Binding.NO_SUPERINTERFACES) {
                            int itsLength2 = itsInterfaces2.length;
                            if (nextPosition + itsLength2 >= interfacesToVisit.length) {
                                System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength2 + 5], 0, nextPosition);
                            }
                            int a2 = 0;
                        Label_0825:
                            while (a2 < itsLength2) {
                                ReferenceBinding next2 = itsInterfaces2[a2];
                                int b2 = 0;
                                Label_0822: {
                                    while (b2 < nextPosition) {
                                        if (TypeBinding.equalsEquals(next2, interfacesToVisit[b2])) {
                                            ++a2;
                                            continue Label_0825;
                                        }
                                        ++b2;
                                    }
                                    interfacesToVisit[nextPosition++] = next2;
                                    continue;
                                }
                            }
                        }
                    }
                }
                if (ambiguous != null) {
                    return ambiguous;
                }
            }
            if (visibleField != null) {
                return visibleField;
            }
            if (notVisibleField != null) {
                return new ProblemFieldBinding(notVisibleField, currentType, fieldName, 2);
            }
            return null;
        }
    }
    
    public ReferenceBinding findMemberType(char[] typeName, ReferenceBinding enclosingType) {
        if ((enclosingType.tagBits & 0x10000L) != 0x0L) {
            return null;
        }
        ReferenceBinding enclosingSourceType = this.enclosingSourceType();
        PackageBinding currentPackage = this.getCurrentPackage();
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordReference(enclosingType, typeName);
        ReferenceBinding memberType = enclosingType.getMemberType(typeName);
        if (memberType != null) {
            unitScope.recordTypeReference(memberType);
            if ((enclosingSourceType != null && (this.parent != unitScope || (enclosingSourceType.tagBits & 0x40000L) != 0x0L)) ? (!memberType.canBeSeenBy(enclosingType, enclosingSourceType)) : (!memberType.canBeSeenBy(currentPackage))) {
                return new ProblemReferenceBinding(new char[][] { typeName }, memberType, 2);
            }
            return memberType;
        }
        else {
            ReferenceBinding currentType = enclosingType;
            ReferenceBinding[] interfacesToVisit = null;
            int nextPosition = 0;
            ReferenceBinding visibleMemberType = null;
            boolean keepLooking = true;
            ReferenceBinding notVisible = null;
            while (keepLooking) {
                ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
                if (itsInterfaces == null) {
                    ReferenceBinding sourceType = currentType.isParameterizedType() ? ((ParameterizedTypeBinding)currentType).genericType() : currentType;
                    if (sourceType.isHierarchyBeingConnected()) {
                        return null;
                    }
                    ((SourceTypeBinding)sourceType).scope.connectTypeHierarchy();
                    itsInterfaces = currentType.superInterfaces();
                }
                if (itsInterfaces != null && itsInterfaces != Binding.NO_SUPERINTERFACES) {
                    if (interfacesToVisit == null) {
                        interfacesToVisit = itsInterfaces;
                        nextPosition = interfacesToVisit.length;
                    }
                    else {
                        int itsLength = itsInterfaces.length;
                        if (nextPosition + itsLength >= interfacesToVisit.length) {
                            System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
                        }
                        int a = 0;
                    Label_0339:
                        while (a < itsLength) {
                            ReferenceBinding next = itsInterfaces[a];
                            int b = 0;
                            Label_0336: {
                                while (b < nextPosition) {
                                    if (TypeBinding.equalsEquals(next, interfacesToVisit[b])) {
                                        ++a;
                                        continue Label_0339;
                                    }
                                    ++b;
                                }
                                interfacesToVisit[nextPosition++] = next;
                                continue;
                            }
                        }
                    }
                }
                if ((currentType = currentType.superclass()) == null) {
                    break;
                }
                unitScope.recordReference(currentType, typeName);
                if ((memberType = currentType.getMemberType(typeName)) != null) {
                    unitScope.recordTypeReference(memberType);
                    keepLooking = false;
                    if (!((enclosingSourceType != null) ? (!memberType.canBeSeenBy(enclosingType, enclosingSourceType)) : (!memberType.canBeSeenBy(currentPackage)))) {
                        if (visibleMemberType == null) {
                            visibleMemberType = memberType;
                        }
                        else {
                            return new ProblemReferenceBinding(new char[][] { typeName }, visibleMemberType, 3);
                        }
                    }
                    else {
                        notVisible = memberType;
                    }
                }
                else {
                    continue;
                }
            }
            if (interfacesToVisit != null) {
                ProblemReferenceBinding ambiguous = null;
                for (int i = 0; i < nextPosition; ++i) {
                    ReferenceBinding anInterface = interfacesToVisit[i];
                    unitScope.recordReference(anInterface, typeName);
                    if ((memberType = anInterface.getMemberType(typeName)) != null) {
                        unitScope.recordTypeReference(memberType);
                        if (visibleMemberType == null) {
                            visibleMemberType = memberType;
                        }
                        else {
                            ambiguous = new ProblemReferenceBinding(new char[][] { typeName }, visibleMemberType, 3);
                            break;
                        }
                    }
                    else {
                        ReferenceBinding[] itsInterfaces2 = anInterface.superInterfaces();
                        if (itsInterfaces2 != null && itsInterfaces2 != Binding.NO_SUPERINTERFACES) {
                            int itsLength2 = itsInterfaces2.length;
                            if (nextPosition + itsLength2 >= interfacesToVisit.length) {
                                System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength2 + 5], 0, nextPosition);
                            }
                            int a2 = 0;
                        Label_0656:
                            while (a2 < itsLength2) {
                                ReferenceBinding next2 = itsInterfaces2[a2];
                                int b2 = 0;
                                Label_0653: {
                                    while (b2 < nextPosition) {
                                        if (TypeBinding.equalsEquals(next2, interfacesToVisit[b2])) {
                                            ++a2;
                                            continue Label_0656;
                                        }
                                        ++b2;
                                    }
                                    interfacesToVisit[nextPosition++] = next2;
                                    continue;
                                }
                            }
                        }
                    }
                }
                if (ambiguous != null) {
                    return ambiguous;
                }
            }
            if (visibleMemberType != null) {
                return visibleMemberType;
            }
            if (notVisible != null) {
                return new ProblemReferenceBinding(new char[][] { typeName }, notVisible, 2);
            }
            return null;
        }
    }
    
    public MethodBinding findMethod(ReferenceBinding receiverType, char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite, boolean inStaticContext) {
        MethodBinding method = this.findMethod0(receiverType, selector, argumentTypes, invocationSite, inStaticContext);
        if (method != null && method.isValidBinding() && method.isVarargs()) {
            TypeBinding elementType = method.parameters[method.parameters.length - 1].leafComponentType();
            if (elementType instanceof ReferenceBinding && !((ReferenceBinding)elementType).canBeSeenBy(this)) {
                return new ProblemMethodBinding(method, method.selector, invocationSite.genericTypeArguments(), 16);
            }
        }
        return method;
    }
    
    public MethodBinding findMethod0(ReferenceBinding receiverType, char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite, boolean inStaticContext) {
        ReferenceBinding currentType = receiverType;
        boolean receiverTypeIsInterface = receiverType.isInterface();
        ObjectVector found = new ObjectVector(3);
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordTypeReferences(argumentTypes);
        List<TypeBinding> visitedTypes = new ArrayList();
        if (receiverTypeIsInterface) {
            unitScope.recordTypeReference(receiverType);
            MethodBinding[] receiverMethods = receiverType.getMethods(selector, argumentTypes.length);
            if (receiverMethods.length > 0) {
                found.addAll(receiverMethods);
            }
            this.findMethodInSuperInterfaces(receiverType, selector, found, visitedTypes, invocationSite);
            currentType = this.getJavaLangObject();
        }
        long complianceLevel = this.compilerOptions().complianceLevel;
        boolean isCompliant14 = complianceLevel >= 3145728L;
        boolean isCompliant15 = complianceLevel >= 3211264L;
        boolean soureLevel18 = this.compilerOptions().sourceLevel >= 3407872L;
        ReferenceBinding classHierarchyStart = currentType;
        MethodVerifier verifier = this.environment().methodVerifier();
        while (currentType != null) {
            unitScope.recordTypeReference(currentType);
            currentType = (ReferenceBinding)currentType.capture(this, (invocationSite == null) ? 0 : invocationSite.sourceStart(), (invocationSite == null) ? 0 : invocationSite.sourceEnd());
            MethodBinding[] currentMethods = currentType.getMethods(selector, argumentTypes.length);
            int currentLength = currentMethods.length;
            if (currentLength > 0) {
                if (isCompliant14 && (receiverTypeIsInterface || found.size > 0)) {
                    for (int i = 0, l = currentLength; i < l; ++i) {
                        MethodBinding currentMethod = currentMethods[i];
                        if (currentMethod != null) {
                            if (receiverTypeIsInterface && !currentMethod.isPublic()) {
                                --currentLength;
                                currentMethods[i] = null;
                            }
                            else {
                                int j = 0;
                                int max = found.size;
                                while (j < max) {
                                    MethodBinding matchingMethod = (MethodBinding)found.elementAt(j);
                                    MethodBinding matchingOriginal = matchingMethod.original();
                                    MethodBinding currentOriginal = matchingOriginal.findOriginalInheritedMethod(currentMethod);
                                    if (currentOriginal != null && verifier.isParameterSubsignature(matchingOriginal, currentOriginal)) {
                                        if (isCompliant15 && matchingMethod.isBridge() && !currentMethod.isBridge()) {
                                            break;
                                        }
                                        --currentLength;
                                        currentMethods[i] = null;
                                        break;
                                    }
                                    else {
                                        ++j;
                                    }
                                }
                            }
                        }
                    }
                }
                if (currentLength > 0) {
                    if (currentMethods.length == currentLength) {
                        found.addAll(currentMethods);
                    }
                    else {
                        for (int i = 0, max2 = currentMethods.length; i < max2; ++i) {
                            MethodBinding currentMethod = currentMethods[i];
                            if (currentMethod != null) {
                                found.add(currentMethod);
                            }
                        }
                    }
                }
            }
            currentType = currentType.superclass();
        }
        int foundSize = found.size;
        MethodBinding[] candidates = null;
        int candidatesCount = 0;
        MethodBinding problemMethod = null;
        boolean searchForDefaultAbstractMethod = soureLevel18 || (isCompliant14 && !receiverTypeIsInterface && (receiverType.isAbstract() || receiverType.isTypeVariable()));
        if (foundSize > 0) {
            for (int k = 0; k < foundSize; ++k) {
                MethodBinding methodBinding = (MethodBinding)found.elementAt(k);
                MethodBinding compatibleMethod = this.computeCompatibleMethod(methodBinding, argumentTypes, invocationSite);
                if (compatibleMethod != null) {
                    if (compatibleMethod.isValidBinding() || compatibleMethod.problemId() == 23) {
                        if (foundSize == 1 && compatibleMethod.canBeSeenBy(receiverType, invocationSite, this)) {
                            if (searchForDefaultAbstractMethod) {
                                return this.findDefaultAbstractMethod(receiverType, selector, argumentTypes, invocationSite, classHierarchyStart, found, new MethodBinding[] { compatibleMethod });
                            }
                            unitScope.recordTypeReferences(compatibleMethod.thrownExceptions);
                            return compatibleMethod;
                        }
                        else {
                            if (candidatesCount == 0) {
                                candidates = new MethodBinding[foundSize];
                            }
                            candidates[candidatesCount++] = compatibleMethod;
                        }
                    }
                    else if (problemMethod == null) {
                        problemMethod = compatibleMethod;
                    }
                }
            }
        }
        if (candidatesCount == 0) {
            if (problemMethod != null) {
                switch (problemMethod.problemId()) {
                    case 11:
                    case 13: {
                        return problemMethod;
                    }
                    default: {
                        break;
                    }
                }
            }
            MethodBinding interfaceMethod2 = this.findDefaultAbstractMethod(receiverType, selector, argumentTypes, invocationSite, classHierarchyStart, found, (MethodBinding[])null);
            if (interfaceMethod2 != null) {
                if (soureLevel18 && foundSize > 0 && interfaceMethod2.isVarargs() && interfaceMethod2 instanceof ParameterizedGenericMethodBinding) {
                    MethodBinding original = interfaceMethod2.original();
                    for (int i4 = 0; i4 < foundSize; ++i4) {
                        MethodBinding classMethod = (MethodBinding)found.elementAt(i4);
                        if (!classMethod.isAbstract()) {
                            MethodBinding substitute = verifier.computeSubstituteMethod(original, classMethod);
                            if (substitute != null && verifier.isSubstituteParameterSubsignature(classMethod, substitute)) {
                                return new ProblemMethodBinding(interfaceMethod2, selector, argumentTypes, 24);
                            }
                        }
                    }
                    return interfaceMethod2;
                }
                return interfaceMethod2;
            }
            else {
                if (found.size == 0) {
                    return null;
                }
                if (problemMethod != null) {
                    return problemMethod;
                }
                int bestArgMatches = -1;
                MethodBinding bestGuess = (MethodBinding)found.elementAt(0);
                int argLength = argumentTypes.length;
                foundSize = found.size;
                for (int i3 = 0; i3 < foundSize; ++i3) {
                    MethodBinding methodBinding3 = (MethodBinding)found.elementAt(i3);
                    TypeBinding[] params = methodBinding3.parameters;
                    int paramLength = params.length;
                    int argMatches = 0;
                    for (int a = 0; a < argLength; ++a) {
                        TypeBinding arg = argumentTypes[a];
                        int p = (a == 0) ? 0 : (a - 1);
                        while (p < paramLength && p < a + 1) {
                            if (TypeBinding.equalsEquals(params[p], arg)) {
                                ++argMatches;
                                break;
                            }
                            ++p;
                        }
                    }
                    if (argMatches >= bestArgMatches) {
                        if (argMatches == bestArgMatches) {
                            int diff1 = (paramLength < argLength) ? (2 * (argLength - paramLength)) : (paramLength - argLength);
                            int bestLength = bestGuess.parameters.length;
                            int diff2 = (bestLength < argLength) ? (2 * (argLength - bestLength)) : (bestLength - argLength);
                            if (diff1 >= diff2) {
                                continue;
                            }
                        }
                        if (bestGuess == methodBinding3 || !MethodVerifier.doesMethodOverride(bestGuess, methodBinding3, this.environment())) {
                            bestArgMatches = argMatches;
                            bestGuess = methodBinding3;
                        }
                    }
                }
                return new ProblemMethodBinding(bestGuess, bestGuess.selector, argumentTypes, 1);
            }
        }
        else {
            int visiblesCount = 0;
            for (int m = 0; m < candidatesCount; ++m) {
                MethodBinding methodBinding2 = candidates[m];
                if (methodBinding2.canBeSeenBy(receiverType, invocationSite, this)) {
                    if (visiblesCount != m) {
                        candidates[m] = null;
                        candidates[visiblesCount] = methodBinding2;
                    }
                    ++visiblesCount;
                }
            }
            switch (visiblesCount) {
                case 0: {
                    MethodBinding interfaceMethod = this.findDefaultAbstractMethod(receiverType, selector, argumentTypes, invocationSite, classHierarchyStart, found, (MethodBinding[])null);
                    if (interfaceMethod != null) {
                        return interfaceMethod;
                    }
                    MethodBinding candidate = candidates[0];
                    int reason = 2;
                    if (candidate.isStatic() && candidate.declaringClass.isInterface()) {
                        if (soureLevel18) {
                            reason = 20;
                        }
                        else {
                            reason = 29;
                        }
                    }
                    return new ProblemMethodBinding(candidate, candidate.selector, candidate.parameters, reason);
                }
                case 1: {
                    if (searchForDefaultAbstractMethod) {
                        return this.findDefaultAbstractMethod(receiverType, selector, argumentTypes, invocationSite, classHierarchyStart, found, new MethodBinding[] { candidates[0] });
                    }
                    MethodBinding candidate = candidates[0];
                    if (candidate != null) {
                        unitScope.recordTypeReferences(candidate.thrownExceptions);
                        return candidate;
                    }
                    return candidate;
                }
                default: {
                    if (complianceLevel <= 3080192L) {
                        ReferenceBinding declaringClass = candidates[0].declaringClass;
                        return declaringClass.isInterface() ? this.mostSpecificInterfaceMethodBinding(candidates, visiblesCount, invocationSite) : this.mostSpecificClassMethodBinding(candidates, visiblesCount, invocationSite);
                    }
                    if (this.compilerOptions().sourceLevel >= 3211264L) {
                        for (int m = 0; m < visiblesCount; ++m) {
                            MethodBinding candidate = candidates[m];
                            if (candidate.isParameterizedGeneric()) {
                                candidate = candidate.shallowOriginal();
                            }
                            if (candidate.hasSubstitutedParameters()) {
                                int j2 = m + 1;
                                while (j2 < visiblesCount) {
                                    MethodBinding otherCandidate = candidates[j2];
                                    if (otherCandidate.hasSubstitutedParameters() && (otherCandidate == candidate || (TypeBinding.equalsEquals(candidate.declaringClass, otherCandidate.declaringClass) && candidate.areParametersEqual(otherCandidate)))) {
                                        return new ProblemMethodBinding(candidates[m], candidates[m].selector, candidates[m].parameters, 3);
                                    }
                                    ++j2;
                                }
                            }
                        }
                    }
                    if (inStaticContext) {
                        MethodBinding[] staticCandidates = new MethodBinding[visiblesCount];
                        int staticCount = 0;
                        for (int i2 = 0; i2 < visiblesCount; ++i2) {
                            if (candidates[i2].isStatic()) {
                                staticCandidates[staticCount++] = candidates[i2];
                            }
                        }
                        if (staticCount == 1) {
                            return staticCandidates[0];
                        }
                        if (staticCount > 1) {
                            return this.mostSpecificMethodBinding(staticCandidates, staticCount, argumentTypes, invocationSite, receiverType);
                        }
                    }
                    if (visiblesCount != candidates.length) {
                        System.arraycopy(candidates, 0, candidates = new MethodBinding[visiblesCount], 0, visiblesCount);
                    }
                    return searchForDefaultAbstractMethod ? this.findDefaultAbstractMethod(receiverType, selector, argumentTypes, invocationSite, classHierarchyStart, found, candidates) : this.mostSpecificMethodBinding(candidates, visiblesCount, argumentTypes, invocationSite, receiverType);
                }
            }
        }
    }
    
    public MethodBinding findMethodForArray(ArrayBinding receiverType, char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite) {
        TypeBinding leafType = receiverType.leafComponentType();
        if (leafType instanceof ReferenceBinding && !((ReferenceBinding)leafType).canBeSeenBy(this)) {
            return new ProblemMethodBinding(selector, Binding.NO_PARAMETERS, (ReferenceBinding)leafType, 8);
        }
        ReferenceBinding object = this.getJavaLangObject();
        MethodBinding methodBinding = object.getExactMethod(selector, argumentTypes, null);
        if (methodBinding != null) {
            if (argumentTypes == Binding.NO_PARAMETERS) {
                switch (selector[0]) {
                    case 'c': {
                        if (CharOperation.equals(selector, TypeConstants.CLONE)) {
                            return this.environment().computeArrayClone(methodBinding);
                        }
                        break;
                    }
                    case 'g': {
                        if (CharOperation.equals(selector, TypeConstants.GETCLASS) && methodBinding.returnType.isParameterizedType()) {
                            return this.environment().createGetClassMethod(receiverType, methodBinding, this);
                        }
                        break;
                    }
                }
            }
            if (methodBinding.canBeSeenBy(receiverType, invocationSite, this)) {
                return methodBinding;
            }
        }
        methodBinding = this.findMethod(object, selector, argumentTypes, invocationSite, false);
        if (methodBinding == null) {
            return new ProblemMethodBinding(selector, argumentTypes, 26);
        }
        return methodBinding;
    }
    
    protected void findMethodInSuperInterfaces(ReferenceBinding receiverType, char[] selector, ObjectVector found, List<TypeBinding> visitedTypes, InvocationSite invocationSite) {
        ReferenceBinding currentType = receiverType;
        ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
        if (itsInterfaces != null && itsInterfaces != Binding.NO_SUPERINTERFACES) {
            ReferenceBinding[] interfacesToVisit = itsInterfaces;
        Label_0393:
            for (int nextPosition = interfacesToVisit.length, i = 0; i < nextPosition; ++i) {
                currentType = interfacesToVisit[i];
                if (visitedTypes != null) {
                    TypeBinding uncaptured = currentType.uncapture(this);
                    for (TypeBinding visited : visitedTypes) {
                        if (uncaptured.isEquivalentTo(visited)) {
                            continue Label_0393;
                        }
                        continue;
                    }
                    visitedTypes.add(uncaptured);
                }
                this.compilationUnitScope().recordTypeReference(currentType);
                currentType = (ReferenceBinding)currentType.capture(this, (invocationSite == null) ? 0 : invocationSite.sourceStart(), (invocationSite == null) ? 0 : invocationSite.sourceEnd());
                MethodBinding[] currentMethods = currentType.getMethods(selector);
                if (currentMethods.length > 0) {
                    int foundSize = found.size;
                Label_0261:
                    for (int c = 0, l = currentMethods.length; c < l; ++c) {
                        MethodBinding current = currentMethods[c];
                        if (current.canBeSeenBy(receiverType, invocationSite, this)) {
                            if (foundSize > 0) {
                                int f = 0;
                                while (f < foundSize) {
                                    if (current == found.elementAt(f)) {
                                        continue Label_0261;
                                    }
                                    ++f;
                                }
                            }
                            found.add(current);
                        }
                    }
                }
                if ((itsInterfaces = currentType.superInterfaces()) != null && itsInterfaces != Binding.NO_SUPERINTERFACES) {
                    int itsLength = itsInterfaces.length;
                    if (nextPosition + itsLength >= interfacesToVisit.length) {
                        System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
                    }
                    int a = 0;
                Label_0386:
                    while (a < itsLength) {
                        ReferenceBinding next = itsInterfaces[a];
                        int b = 0;
                        Label_0383: {
                            while (b < nextPosition) {
                                if (TypeBinding.equalsEquals(next, interfacesToVisit[b])) {
                                    ++a;
                                    continue Label_0386;
                                }
                                ++b;
                            }
                            interfacesToVisit[nextPosition++] = next;
                            continue;
                        }
                    }
                }
            }
        }
    }
    
    public ReferenceBinding findType(char[] typeName, PackageBinding declarationPackage, PackageBinding invocationPackage) {
        this.compilationUnitScope().recordReference(declarationPackage.compoundName, typeName);
        ReferenceBinding typeBinding = declarationPackage.getType(typeName);
        if (typeBinding == null) {
            return null;
        }
        if (typeBinding.isValidBinding() && declarationPackage != invocationPackage && !typeBinding.canBeSeenBy(invocationPackage)) {
            return new ProblemReferenceBinding(new char[][] { typeName }, typeBinding, 2);
        }
        return typeBinding;
    }
    
    public LocalVariableBinding findVariable(char[] variable) {
        return null;
    }
    
    public Binding getBinding(char[] name, int mask, InvocationSite invocationSite, boolean needResolve) {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        LookupEnvironment env = unitScope.environment;
        try {
            env.missingClassFileLocation = invocationSite;
            Binding binding = null;
            FieldBinding problemField = null;
            Label_1177: {
                if ((mask & 0x3) != 0x0) {
                    boolean insideStaticContext = false;
                    boolean insideConstructorCall = false;
                    boolean insideTypeAnnotation = false;
                    FieldBinding foundField = null;
                    ProblemFieldBinding foundInsideProblem = null;
                    Scope scope = this;
                    MethodScope methodScope = null;
                    int depth = 0;
                    int foundDepth = 0;
                    boolean shouldTrackOuterLocals = false;
                    ReferenceBinding foundActualReceiverType = null;
                    while (true) {
                        switch (scope.kind) {
                            case 2: {
                                methodScope = (MethodScope)scope;
                                insideStaticContext |= methodScope.isStatic;
                                insideConstructorCall |= methodScope.isConstructorCall;
                                insideTypeAnnotation = methodScope.insideTypeAnnotation;
                            }
                            case 1: {
                                LocalVariableBinding variableBinding = scope.findVariable(name);
                                if (variableBinding != null) {
                                    if (foundField != null && foundField.isValidBinding()) {
                                        return new ProblemFieldBinding(foundField, foundField.declaringClass, name, 5);
                                    }
                                    if (depth > 0) {
                                        invocationSite.setDepth(depth);
                                    }
                                    if (shouldTrackOuterLocals) {
                                        if ((invocationSite instanceof NameReference)) {
                                            NameReference nameReference = (NameReference)invocationSite;
                                            nameReference.bits |= 0x80000;
                                        }
                                        else if ((invocationSite instanceof AbstractVariableDeclaration)) {
                                            AbstractVariableDeclaration variableDeclaration = (AbstractVariableDeclaration)invocationSite;
                                            variableDeclaration.bits |= 0x200000;
                                        }
                                    }
                                    return variableBinding;
                                }
                                else {
                                    break;
                                }
                            }
                            case 3: {
                                ClassScope classScope = (ClassScope)scope;
                                ReferenceBinding receiverType = classScope.enclosingReceiverType();
                                if (!insideTypeAnnotation) {
                                    FieldBinding fieldBinding = classScope.findField(receiverType, name, invocationSite, needResolve);
                                    if (fieldBinding != null) {
                                        if (fieldBinding.problemId() == 3) {
                                            if (foundField == null || foundField.problemId() == 2) {
                                                return fieldBinding;
                                            }
                                            return new ProblemFieldBinding(foundField, foundField.declaringClass, name, 5);
                                        }
                                        else {
                                            ProblemFieldBinding insideProblem = null;
                                            if (fieldBinding.isValidBinding()) {
                                                if (!fieldBinding.isStatic()) {
                                                    if (insideConstructorCall) {
                                                        insideProblem = new ProblemFieldBinding(fieldBinding, fieldBinding.declaringClass, name, 6);
                                                    }
                                                    else if (insideStaticContext) {
                                                        insideProblem = new ProblemFieldBinding(fieldBinding, fieldBinding.declaringClass, name, 7);
                                                    }
                                                }
                                                if (TypeBinding.equalsEquals(receiverType, fieldBinding.declaringClass) || this.compilerOptions().complianceLevel >= 3145728L) {
                                                    if (foundField == null) {
                                                        if (depth > 0) {
                                                            invocationSite.setDepth(depth);
                                                            invocationSite.setActualReceiverType(receiverType);
                                                        }
                                                        insideProblem = ((insideProblem == null) ? fieldBinding : insideProblem);
                                                        return insideProblem;
                                                    }
                                                    else if (foundField.isValidBinding() && TypeBinding.notEquals(foundField.declaringClass, fieldBinding.declaringClass) && TypeBinding.notEquals(foundField.declaringClass, foundActualReceiverType)) {
                                                        return new ProblemFieldBinding(foundField, foundField.declaringClass, name, 5);
                                                    }
                                                }
                                            }
                                            if (foundField == null || (foundField.problemId() == 2 && fieldBinding.problemId() != 2)) {
                                                foundDepth = depth;
                                                foundActualReceiverType = receiverType;
                                                foundInsideProblem = insideProblem;
                                                foundField = fieldBinding;
                                            }
                                        }
                                    }
                                }
                                insideTypeAnnotation = false;
                                ++depth;
                                shouldTrackOuterLocals = true;
                                insideStaticContext |= receiverType.isStatic();
                                MethodScope enclosingMethodScope = scope.methodScope();
                                insideConstructorCall = (enclosingMethodScope != null && enclosingMethodScope.isConstructorCall);
                                break;
                            }
                            case 4: {
                                if (foundInsideProblem != null) {
                                    return foundInsideProblem;
                                }
                                if (foundField != null) {
                                    if (foundField.isValidBinding()) {
                                        if (foundDepth > 0) {
                                            invocationSite.setDepth(foundDepth);
                                            invocationSite.setActualReceiverType(foundActualReceiverType);
                                        }
                                        return foundField;
                                    }
                                    else {
                                        problemField = foundField;
                                        foundField = null;
                                    }
                                }
                                if (this.compilerOptions().sourceLevel >= 3211264L) {
                                    unitScope.faultInImports();
                                    ImportBinding[] imports = unitScope.imports;
                                    if (imports != null) {
                                        for (int i = 0, length = imports.length; i < length; ++i) {
                                            ImportBinding importBinding = imports[i];
                                            if (importBinding.isStatic() && !importBinding.onDemand && CharOperation.equals(importBinding.compoundName[importBinding.compoundName.length - 1], name) && unitScope.resolveSingleImport(importBinding, 13) != null && importBinding.resolvedImport instanceof FieldBinding) {
                                                foundField = (FieldBinding)importBinding.resolvedImport;
                                                ImportReference importReference = importBinding.reference;
                                                if (importReference != null && needResolve) {
                                                    importReference.bits |= 0x2;
                                                }
                                                invocationSite.setActualReceiverType(foundField.declaringClass);
                                                if (foundField.isValidBinding()) {
                                                    return foundField;
                                                }
                                                if (problemField == null) {
                                                    problemField = foundField;
                                                }
                                            }
                                        }
                                        boolean foundInImport = false;
                                        for (int j = 0, length2 = imports.length; j < length2; ++j) {
                                            ImportBinding importBinding2 = imports[j];
                                            if (importBinding2.isStatic() && importBinding2.onDemand) {
                                                Binding resolvedImport = importBinding2.resolvedImport;
                                                if ((resolvedImport instanceof ReferenceBinding)) {
                                                    FieldBinding temp = this.findField((ReferenceBinding)resolvedImport, name, invocationSite, needResolve);
                                                    if (temp != null) {
                                                        if (!temp.isValidBinding()) {
                                                            if (problemField == null) {
                                                                problemField = temp;
                                                            }
                                                        }
                                                        else if (temp.isStatic() && foundField != temp) {
                                                            ImportReference importReference2 = importBinding2.reference;
                                                            if (importReference2 != null && needResolve) {
                                                                importReference2.bits |= 0x2;
                                                            }
                                                            if (foundInImport) {
                                                                return new ProblemFieldBinding(foundField, foundField.declaringClass, name, 3);
                                                            }
                                                            foundField = temp;
                                                            foundInImport = true;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        if (foundField != null) {
                                            invocationSite.setActualReceiverType(foundField.declaringClass);
                                            return foundField;
                                        }
                                        break Label_1177;
                                    }
                                    else {
                                        break Label_1177;
                                    }
                                }
                                else {
                                    break Label_1177;
                                }
                            }
                        }
                        if (scope.isLambdaScope()) {
                            shouldTrackOuterLocals = true;
                        }
                        scope = scope.parent;
                    }
                }
            }
            if ((mask & 0x4) != 0x0) {
                if ((binding = getBaseType(name)) != null) {
                    return binding;
                }
                binding = this.getTypeOrPackage(name, ((mask & 0x10) == 0x0) ? 4 : 20, needResolve);
                if (binding.isValidBinding() || mask == 4) {
                    return binding;
                }
            }
            else if ((mask & 0x10) != 0x0) {
                unitScope.recordSimpleReference(name);
                if ((binding = env.getTopLevelPackage(name)) != null) {
                    return binding;
                }
            }
            if (problemField != null) {
                return problemField;
            }
            if (binding != null && binding.problemId() != 1) {
                return binding;
            }
            return new ProblemBinding(name, this.enclosingSourceType(), 1);
        }
        catch (AbortCompilation e) {
            e.updateContext(invocationSite, this.referenceCompilationUnit().compilationResult);
            throw e;
        }
        finally {
            env.missingClassFileLocation = null;
        }
    }
    
    private MethodBinding getExactMethod(TypeBinding receiverType, TypeBinding type, char[] selector, InvocationSite invocationSite, MethodBinding candidate) {
        if (type == null) {
            return null;
        }
        TypeBinding[] superInterfaces = type.superInterfaces();
        TypeBinding[] typePlusSupertypes = new TypeBinding[2 + superInterfaces.length];
        typePlusSupertypes[0] = type;
        typePlusSupertypes[1] = type.superclass();
        if (superInterfaces.length != 0) {
            System.arraycopy(superInterfaces, 0, typePlusSupertypes, 2, superInterfaces.length);
        }
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordTypeReference(type);
        type = type.capture(this, invocationSite.sourceStart(), invocationSite.sourceEnd());
        for (int i = 0, typesLength = typePlusSupertypes.length; i < typesLength; ++i) {
            MethodBinding[] methods = (i == 0) ? type.getMethods(selector) : new MethodBinding[] { this.getExactMethod(receiverType, typePlusSupertypes[i], selector, invocationSite, candidate) };
            for (int j = 0, length = methods.length; j < length; ++j) {
                MethodBinding currentMethod = methods[j];
                if (currentMethod != null && candidate != currentMethod) {
                    if (i == 0) {
                        if (currentMethod.canBeSeenBy(receiverType, invocationSite, this) && !currentMethod.isSynthetic()) {
                            if (currentMethod.isBridge()) {
                                continue;
                            }
                        }
                        else {
                            continue;
                        }
                    }
                    if (candidate != null) {
                        if (!candidate.areParameterErasuresEqual(currentMethod)) {
                            throw new Scope.MethodClashException();
                        }
                    }
                    else {
                        candidate = currentMethod;
                    }
                }
            }
        }
        return candidate;
    }
    
    public MethodBinding getExactMethod(TypeBinding receiverType, char[] selector, InvocationSite invocationSite) {
        if (receiverType == null || !receiverType.isValidBinding() || receiverType.isBaseType()) {
            return null;
        }
        TypeBinding currentType = receiverType;
        if (currentType.isArrayType()) {
            if (!currentType.leafComponentType().canBeSeenBy(this)) {
                return null;
            }
            currentType = this.getJavaLangObject();
        }
        MethodBinding exactMethod = null;
        try {
            exactMethod = this.getExactMethod(receiverType, currentType, selector, invocationSite, null);
        }
        catch (Scope.MethodClashException ex) {
            return null;
        }
        if (exactMethod == null || !exactMethod.canBeSeenBy(invocationSite, this)) {
            return null;
        }
        TypeBinding[] typeArguments = invocationSite.genericTypeArguments();
        TypeVariableBinding[] typeVariables = exactMethod.typeVariables();
        if (exactMethod.isVarargs() || (typeVariables != Binding.NO_TYPE_VARIABLES && (typeArguments == null || typeArguments.length != typeVariables.length))) {
            return null;
        }
        if (receiverType.isArrayType()) {
            if (CharOperation.equals(selector, TypeConstants.CLONE)) {
                return this.environment().computeArrayClone(exactMethod);
            }
            if (CharOperation.equals(selector, TypeConstants.GETCLASS)) {
                return this.environment().createGetClassMethod(receiverType, exactMethod, this);
            }
        }
        if (exactMethod.declaringClass.id == 1 && CharOperation.equals(selector, TypeConstants.GETCLASS) && exactMethod.returnType.isParameterizedType()) {
            return this.environment().createGetClassMethod(receiverType, exactMethod, this);
        }
        if (typeVariables != Binding.NO_TYPE_VARIABLES) {
            return this.environment().createParameterizedGenericMethod(exactMethod, typeArguments);
        }
        return exactMethod;
    }
    
    public MethodBinding getExactConstructor(TypeBinding receiverType, InvocationSite invocationSite) {
        if (receiverType == null || !receiverType.isValidBinding() || !receiverType.canBeInstantiated() || receiverType.isBaseType()) {
            return null;
        }
        if (receiverType.isArrayType()) {
            TypeBinding leafType = receiverType.leafComponentType();
            if (!leafType.canBeSeenBy(this) || !leafType.isReifiable()) {
                return null;
            }
            return new MethodBinding(4097, TypeConstants.INIT, receiverType, new TypeBinding[] { TypeBinding.INT }, Binding.NO_EXCEPTIONS, this.getJavaLangObject());
        }
        else {
            CompilationUnitScope unitScope = this.compilationUnitScope();
            MethodBinding exactConstructor = null;
            unitScope.recordTypeReference(receiverType);
            MethodBinding[] methods = receiverType.getMethods(TypeConstants.INIT);
            TypeBinding[] genericTypeArguments = invocationSite.genericTypeArguments();
            for (int i = 0, length = methods.length; i < length; ++i) {
                MethodBinding constructor = methods[i];
                if (constructor.canBeSeenBy(invocationSite, this)) {
                    if (constructor.isVarargs()) {
                        return null;
                    }
                    if (constructor.typeVariables() != Binding.NO_TYPE_VARIABLES && genericTypeArguments == null) {
                        return null;
                    }
                    if (exactConstructor == null) {
                        exactConstructor = constructor;
                    }
                    else {
                        return null;
                    }
                }
            }
            if (exactConstructor != null) {
                TypeVariableBinding[] typeVariables = exactConstructor.typeVariables();
                if (typeVariables != Binding.NO_TYPE_VARIABLES) {
                    if (typeVariables.length != genericTypeArguments.length) {
                        return null;
                    }
                    exactConstructor = this.environment().createParameterizedGenericMethod(exactConstructor, genericTypeArguments);
                }
            }
            return exactConstructor;
        }
    }
    
    public MethodBinding getConstructor(ReferenceBinding receiverType, TypeBinding[] argumentTypes, InvocationSite invocationSite) {
        MethodBinding method = this.getConstructor0(receiverType, argumentTypes, invocationSite);
        if (method != null && method.isValidBinding() && method.isVarargs()) {
            TypeBinding elementType = method.parameters[method.parameters.length - 1].leafComponentType();
            if (elementType instanceof ReferenceBinding && !((ReferenceBinding)elementType).canBeSeenBy(this)) {
                return new ProblemMethodBinding(method, method.selector, invocationSite.genericTypeArguments(), 16);
            }
        }
        return method;
    }
    
    public MethodBinding getConstructor0(ReferenceBinding receiverType, TypeBinding[] argumentTypes, InvocationSite invocationSite) {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        LookupEnvironment env = unitScope.environment;
        try {
            env.missingClassFileLocation = invocationSite;
            unitScope.recordTypeReference(receiverType);
            unitScope.recordTypeReferences(argumentTypes);
            MethodBinding methodBinding = receiverType.getExactConstructor(argumentTypes);
            if (methodBinding != null && methodBinding.canBeSeenBy(invocationSite, this)) {
                if (invocationSite.genericTypeArguments() != null) {
                    methodBinding = this.computeCompatibleMethod(methodBinding, argumentTypes, invocationSite);
                }
                return methodBinding;
            }
            else {
                MethodBinding[] methods = receiverType.getMethods(TypeConstants.INIT, argumentTypes.length);
                if (methods == Binding.NO_METHODS) {
                    return new ProblemMethodBinding(TypeConstants.INIT, argumentTypes, 1);
                }
                MethodBinding[] compatible = new MethodBinding[methods.length];
                int compatibleIndex = 0;
                MethodBinding problemMethod = null;
                for (int i = 0, length = methods.length; i < length; ++i) {
                    MethodBinding compatibleMethod = this.computeCompatibleMethod(methods[i], argumentTypes, invocationSite);
                    if (compatibleMethod != null) {
                        if (compatibleMethod.isValidBinding()) {
                            compatible[compatibleIndex++] = compatibleMethod;
                        }
                        else if (problemMethod == null) {
                            problemMethod = compatibleMethod;
                        }
                    }
                }
                if (compatibleIndex == 0) {
                    if (problemMethod == null) {
                        return new ProblemMethodBinding(methods[0], TypeConstants.INIT, argumentTypes, 1);
                    }
                    return problemMethod;
                }
                else {
                    MethodBinding[] visible = new MethodBinding[compatibleIndex];
                    int visibleIndex = 0;
                    for (MethodBinding method : compatible) {
                        if (method.canBeSeenBy(invocationSite, this)) {
                            visible[visibleIndex++] = method;
                        }
                    }
                    if (visibleIndex == 1) {
                        return visible[0];
                    }
                    if (visibleIndex == 0) {
                        return new ProblemMethodBinding(compatible[0], TypeConstants.INIT, compatible[0].parameters, 2);
                    }
                    return this.mostSpecificMethodBinding(visible, visibleIndex, argumentTypes, invocationSite, receiverType);
                }
            }
        }
        catch (AbortCompilation e) {
            e.updateContext(invocationSite, this.referenceCompilationUnit().compilationResult);
            throw e;
        }
        finally {
            env.missingClassFileLocation = null;
        }
    }
    
    public final PackageBinding getCurrentPackage() {
        Scope unitScope;
        Scope scope;
        for (unitScope = this; (scope = unitScope.parent) != null; unitScope = scope) {}
        return ((CompilationUnitScope)unitScope).fPackage;
    }
    
    public int getDeclarationModifiers() {
        switch (this.kind) {
            case 1:
            case 2: {
                MethodScope methodScope = this.methodScope();
                if (!methodScope.isInsideInitializer()) {
                    MethodBinding context = ((AbstractMethodDeclaration)methodScope.referenceContext).binding;
                    if (context != null) {
                        return context.modifiers;
                    }
                    break;
                }
                else {
                    SourceTypeBinding type = ((BlockScope)this).referenceType().binding;
                    if (methodScope.initializedField != null) {
                        return methodScope.initializedField.modifiers;
                    }
                    if (type != null) {
                        return type.modifiers;
                    }
                    break;
                }
            }
            case 3: {
                ReferenceBinding context2 = ((ClassScope)this).referenceType().binding;
                if (context2 != null) {
                    return context2.modifiers;
                }
                break;
            }
        }
        return -1;
    }
    
    public FieldBinding getField(TypeBinding receiverType, char[] fieldName, InvocationSite invocationSite) {
        LookupEnvironment env = this.environment();
        try {
            env.missingClassFileLocation = invocationSite;
            FieldBinding field = this.findField(receiverType, fieldName, invocationSite, true);
            if (field != null) {
                return field;
            }
            return new ProblemFieldBinding((receiverType instanceof ReferenceBinding) ? ((ReferenceBinding)receiverType) : null, fieldName, 1);
        }
        catch (AbortCompilation e) {
            e.updateContext(invocationSite, this.referenceCompilationUnit().compilationResult);
            throw e;
        }
        finally {
            env.missingClassFileLocation = null;
        }
    }
    
    public MethodBinding getImplicitMethod(char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite) {
        boolean insideStaticContext = false;
        boolean insideConstructorCall = false;
        boolean insideTypeAnnotation = false;
        MethodBinding foundMethod = null;
        MethodBinding foundProblem = null;
        boolean foundProblemVisible = false;
        Scope scope = this;
        MethodScope methodScope = null;
        int depth = 0;
        CompilerOptions options;
        boolean inheritedHasPrecedence = (options = this.compilerOptions()).complianceLevel >= 3145728L;
        while (true) {
            switch (scope.kind) {
                case 2: {
                    methodScope = (MethodScope)scope;
                    insideStaticContext |= methodScope.isStatic;
                    insideConstructorCall |= methodScope.isConstructorCall;
                    insideTypeAnnotation = methodScope.insideTypeAnnotation;
                    break;
                }
                case 3: {
                    ClassScope classScope = (ClassScope)scope;
                    ReferenceBinding receiverType = classScope.enclosingReceiverType();
                    if (!insideTypeAnnotation) {
                        MethodBinding methodBinding = classScope.findExactMethod(receiverType, selector, argumentTypes, invocationSite);
                        if (methodBinding == null) {
                            methodBinding = classScope.findMethod(receiverType, selector, argumentTypes, invocationSite, false);
                        }
                        if (methodBinding != null) {
                            if (foundMethod == null) {
                                if (methodBinding.isValidBinding()) {
                                    if (!methodBinding.isStatic() && (insideConstructorCall || insideStaticContext)) {
                                        if (foundProblem != null && foundProblem.problemId() != 2) {
                                            return foundProblem;
                                        }
                                        return new ProblemMethodBinding(methodBinding, methodBinding.selector, methodBinding.parameters, insideConstructorCall ? 6 : 7);
                                    }
                                    else {
                                        if (!methodBinding.isStatic() && methodScope != null) {
                                            this.tagAsAccessingEnclosingInstanceStateOf(receiverType, false);
                                        }
                                        if (inheritedHasPrecedence || TypeBinding.equalsEquals(receiverType, methodBinding.declaringClass) || receiverType.getMethods(selector) != Binding.NO_METHODS) {
                                            if (foundProblemVisible) {
                                                return foundProblem;
                                            }
                                            if (depth > 0) {
                                                invocationSite.setDepth(depth);
                                                invocationSite.setActualReceiverType(receiverType);
                                            }
                                            if (argumentTypes == Binding.NO_PARAMETERS && CharOperation.equals(selector, TypeConstants.GETCLASS) && methodBinding.returnType.isParameterizedType()) {
                                                return this.environment().createGetClassMethod(receiverType, methodBinding, this);
                                            }
                                            return methodBinding;
                                        }
                                        else if (foundProblem == null || foundProblem.problemId() == 2) {
                                            if (foundProblem != null) {
                                                foundProblem = null;
                                            }
                                            if (depth > 0) {
                                                invocationSite.setDepth(depth);
                                                invocationSite.setActualReceiverType(receiverType);
                                            }
                                            foundMethod = methodBinding;
                                        }
                                    }
                                }
                                else {
                                    if (methodBinding.problemId() != 2 && methodBinding.problemId() != 1) {
                                        return methodBinding;
                                    }
                                    if (foundProblem == null) {
                                        foundProblem = methodBinding;
                                    }
                                    if (!foundProblemVisible && methodBinding.problemId() == 1) {
                                        MethodBinding closestMatch = ((ProblemMethodBinding)methodBinding).closestMatch;
                                        if (closestMatch != null && closestMatch.canBeSeenBy(receiverType, invocationSite, this)) {
                                            foundProblem = methodBinding;
                                            foundProblemVisible = true;
                                        }
                                    }
                                }
                            }
                            else if (methodBinding.problemId() == 3 || (TypeBinding.notEquals(foundMethod.declaringClass, methodBinding.declaringClass) && (TypeBinding.equalsEquals(receiverType, methodBinding.declaringClass) || receiverType.getMethods(selector) != Binding.NO_METHODS))) {
                                return new ProblemMethodBinding(methodBinding, selector, argumentTypes, 5);
                            }
                        }
                    }
                    insideTypeAnnotation = false;
                    ++depth;
                    insideStaticContext |= receiverType.isStatic();
                    MethodScope enclosingMethodScope = scope.methodScope();
                    insideConstructorCall = (enclosingMethodScope != null && enclosingMethodScope.isConstructorCall);
                    break;
                }
                case 4: {
                    if (insideStaticContext && options.sourceLevel >= 3211264L) {
                        if (foundProblem != null) {
                            if (foundProblem.declaringClass != null && foundProblem.declaringClass.id == 1) {
                                return foundProblem;
                            }
                            if (foundProblem.problemId() == 1 && foundProblemVisible) {
                                return foundProblem;
                            }
                        }
                        CompilationUnitScope unitScope = (CompilationUnitScope)scope;
                        unitScope.faultInImports();
                        ImportBinding[] imports = unitScope.imports;
                        if (imports != null) {
                            ObjectVector visible = null;
                            boolean skipOnDemand = false;
                            for (int i = 0, length = imports.length; i < length; ++i) {
                                ImportBinding importBinding = imports[i];
                                if (importBinding.isStatic()) {
                                    Binding resolvedImport = importBinding.resolvedImport;
                                    MethodBinding possible = null;
                                    if (importBinding.onDemand) {
                                        if (!skipOnDemand && resolvedImport instanceof ReferenceBinding) {
                                            possible = this.findMethod((ReferenceBinding)resolvedImport, selector, argumentTypes, invocationSite, true);
                                        }
                                    }
                                    else if ((resolvedImport instanceof MethodBinding)) {
                                        MethodBinding staticMethod = (MethodBinding)resolvedImport;
                                        if (CharOperation.equals(staticMethod.selector, selector)) {
                                            possible = this.findMethod(staticMethod.declaringClass, selector, argumentTypes, invocationSite, true);
                                        }
                                    }
                                    else if ((resolvedImport instanceof FieldBinding)) {
                                        FieldBinding staticField = (FieldBinding)resolvedImport;
                                        if (CharOperation.equals(staticField.name, selector)) {
                                            char[][] importName = importBinding.reference.tokens;
                                            TypeBinding referencedType = this.getType(importName, importName.length - 1);
                                            if (referencedType != null) {
                                                possible = this.findMethod((ReferenceBinding)referencedType, selector, argumentTypes, invocationSite, true);
                                            }
                                        }
                                    }
                                    if (possible != null && possible != foundProblem) {
                                        if (!possible.isValidBinding()) {
                                            if (foundProblem == null) {
                                                foundProblem = possible;
                                            }
                                        }
                                        else if (possible.isStatic()) {
                                            MethodBinding compatibleMethod = this.computeCompatibleMethod(possible, argumentTypes, invocationSite);
                                            if (compatibleMethod != null) {
                                                if (compatibleMethod.isValidBinding()) {
                                                    if (compatibleMethod.canBeSeenBy(unitScope.fPackage)) {
                                                        if (!skipOnDemand && !importBinding.onDemand) {
                                                            visible = null;
                                                            skipOnDemand = true;
                                                        }
                                                        if (visible == null || !visible.contains(compatibleMethod)) {
                                                            ImportReference importReference = importBinding.reference;
                                                            if (importReference != null) {
                                                                importReference.bits |= 0x2;
                                                            }
                                                            if (visible == null) {
                                                                visible = new ObjectVector(3);
                                                            }
                                                            visible.add(compatibleMethod);
                                                        }
                                                    }
                                                    else if (foundProblem == null) {
                                                        foundProblem = new ProblemMethodBinding(compatibleMethod, selector, compatibleMethod.parameters, 2);
                                                    }
                                                }
                                                else if (foundProblem == null) {
                                                    foundProblem = compatibleMethod;
                                                }
                                            }
                                            else if (foundProblem == null) {
                                                foundProblem = new ProblemMethodBinding(possible, selector, argumentTypes, 1);
                                            }
                                        }
                                    }
                                }
                            }
                            if (visible != null) {
                                if (visible.size == 1) {
                                    foundMethod = (MethodBinding)visible.elementAt(0);
                                }
                                else {
                                    MethodBinding[] temp = new MethodBinding[visible.size];
                                    visible.copyInto(temp);
                                    foundMethod = this.mostSpecificMethodBinding(temp, temp.length, argumentTypes, invocationSite, null);
                                }
                            }
                        }
                    }
                    if (foundMethod != null) {
                        invocationSite.setActualReceiverType(foundMethod.declaringClass);
                        return foundMethod;
                    }
                    if (foundProblem != null) {
                        return foundProblem;
                    }
                    return new ProblemMethodBinding(selector, argumentTypes, 1);
                }
            }
            scope = scope.parent;
        }
    }
    
    public final ReferenceBinding getJavaIoSerializable() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_IO_SERIALIZABLE);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_IO_SERIALIZABLE, this);
    }
    
    public final ReferenceBinding getJavaLangAnnotationAnnotation() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_ANNOTATION_ANNOTATION);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_ANNOTATION_ANNOTATION, this);
    }
    
    public final ReferenceBinding getJavaLangAssertionError() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_ASSERTIONERROR);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_ASSERTIONERROR, this);
    }
    
    public final ReferenceBinding getJavaLangClass() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_CLASS);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_CLASS, this);
    }
    
    public final ReferenceBinding getJavaLangCloneable() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_CLONEABLE);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_CLONEABLE, this);
    }
    
    public final ReferenceBinding getJavaLangEnum() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_ENUM);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_ENUM, this);
    }
    
    public final ReferenceBinding getJavaLangInvokeLambdaMetafactory() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_INVOKE_LAMBDAMETAFACTORY);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_INVOKE_LAMBDAMETAFACTORY, this);
    }
    
    public final ReferenceBinding getJavaLangInvokeSerializedLambda() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_INVOKE_SERIALIZEDLAMBDA);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_INVOKE_SERIALIZEDLAMBDA, this);
    }
    
    public final ReferenceBinding getJavaLangInvokeMethodHandlesLookup() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_INVOKE_METHODHANDLES);
        ReferenceBinding outerType = unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_INVOKE_METHODHANDLES, this);
        return this.findDirectMemberType("Lookup".toCharArray(), outerType);
    }
    
    public final ReferenceBinding getJavaLangIterable() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_ITERABLE);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_ITERABLE, this);
    }
    
    public final ReferenceBinding getJavaLangObject() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_OBJECT);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_OBJECT, this);
    }
    
    public final ReferenceBinding getJavaLangString() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_STRING);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_STRING, this);
    }
    
    public final ReferenceBinding getJavaLangThrowable() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_THROWABLE);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_THROWABLE, this);
    }
    
    public final ReferenceBinding getJavaLangIllegalArgumentException() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_LANG_ILLEGALARGUMENTEXCEPTION);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_LANG_ILLEGALARGUMENTEXCEPTION, this);
    }
    
    public final ReferenceBinding getJavaUtilIterator() {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(TypeConstants.JAVA_UTIL_ITERATOR);
        return unitScope.environment.getResolvedType(TypeConstants.JAVA_UTIL_ITERATOR, this);
    }
    
    public final ReferenceBinding getMemberType(char[] typeName, ReferenceBinding enclosingType) {
        ReferenceBinding memberType = this.findMemberType(typeName, enclosingType);
        if (memberType != null) {
            return memberType;
        }
        char[][] compoundName = { typeName };
        return new ProblemReferenceBinding(compoundName, null, 1);
    }
    
    public MethodBinding getMethod(TypeBinding receiverType, char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite) {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        LookupEnvironment env = unitScope.environment;
        try {
            env.missingClassFileLocation = invocationSite;
            switch (receiverType.kind()) {
                case 132: {
                    return new ProblemMethodBinding(selector, argumentTypes, 1);
                }
                case 68: {
                    unitScope.recordTypeReference(receiverType);
                    return this.findMethodForArray((ArrayBinding)receiverType, selector, argumentTypes, invocationSite);
                }
                default: {
                    unitScope.recordTypeReference(receiverType);
                    ReferenceBinding currentType = (ReferenceBinding)receiverType;
                    if (!currentType.canBeSeenBy(this)) {
                        return new ProblemMethodBinding(selector, argumentTypes, 8);
                    }
                    MethodBinding methodBinding = this.findExactMethod(currentType, selector, argumentTypes, invocationSite);
                    if (methodBinding != null && methodBinding.isValidBinding()) {
                        return methodBinding;
                    }
                    methodBinding = this.findMethod(currentType, selector, argumentTypes, invocationSite, false);
                    if (methodBinding == null) {
                        return new ProblemMethodBinding(selector, argumentTypes, 1);
                    }
                    if (!methodBinding.isValidBinding()) {
                        return methodBinding;
                    }
                    if (argumentTypes == Binding.NO_PARAMETERS && CharOperation.equals(selector, TypeConstants.GETCLASS) && methodBinding.returnType.isParameterizedType()) {
                        return this.environment().createGetClassMethod(receiverType, methodBinding, this);
                    }
                    return methodBinding;
                }
            }
        }
        catch (AbortCompilation e) {
            e.updateContext(invocationSite, this.referenceCompilationUnit().compilationResult);
            throw e;
        }
        finally {
            env.missingClassFileLocation = null;
        }
    }
    
    public final Binding getPackage(char[][] compoundName) {
        this.compilationUnitScope().recordQualifiedReference(compoundName);
        Binding binding = this.getTypeOrPackage(compoundName[0], 20, true);
        if (binding == null) {
            char[][] qName = { compoundName[0] };
            return new ProblemReferenceBinding(qName, this.environment().createMissingType(null, compoundName), 1);
        }
        if (!binding.isValidBinding()) {
            if ((binding instanceof PackageBinding)) {
                char[][] qName = { compoundName[0] };
                return new ProblemReferenceBinding(qName, null, 1);
            }
            return binding;
        }
        else {
            if (!(binding instanceof PackageBinding)) {
                return null;
            }
            int currentIndex = 1;
            int length = compoundName.length;
            PackageBinding packageBinding = (PackageBinding)binding;
            while (currentIndex < length) {
                binding = packageBinding.getTypeOrPackage(compoundName[currentIndex++]);
                if (binding == null) {
                    return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), null, 1);
                }
                if (!binding.isValidBinding()) {
                    return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), (binding instanceof ReferenceBinding) ? ((ReferenceBinding)((ReferenceBinding)binding).closestMatch()) : null, binding.problemId());
                }
                if (!(binding instanceof PackageBinding)) {
                    return packageBinding;
                }
                packageBinding = (PackageBinding)binding;
            }
            return new ProblemReferenceBinding(compoundName, null, 1);
        }
    }
    
    public final Binding getOnlyPackage(char[][] compoundName) {
        this.compilationUnitScope().recordQualifiedReference(compoundName);
        Binding binding = this.getTypeOrPackage(compoundName[0], 16, true);
        if (binding == null || !binding.isValidBinding()) {
            char[][] qName = { compoundName[0] };
            return new ProblemReferenceBinding(qName, null, 1);
        }
        if (!(binding instanceof PackageBinding)) {
            return null;
        }
        int currentIndex = 1;
        int length = compoundName.length;
        PackageBinding packageBinding = (PackageBinding)binding;
        while (currentIndex < length) {
            binding = packageBinding.getPackage(compoundName[currentIndex++]);
            if (binding == null) {
                return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), null, 1);
            }
            if (!binding.isValidBinding()) {
                return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), (binding instanceof ReferenceBinding) ? ((ReferenceBinding)((ReferenceBinding)binding).closestMatch()) : null, binding.problemId());
            }
            packageBinding = (PackageBinding)binding;
        }
        return packageBinding;
    }
    
    public final TypeBinding getType(char[] name) {
        TypeBinding binding = getBaseType(name);
        if (binding != null) {
            return binding;
        }
        return (ReferenceBinding)this.getTypeOrPackage(name, 4, true);
    }
    
    public final TypeBinding getType(char[] name, PackageBinding packageBinding) {
        if (packageBinding == null) {
            return this.getType(name);
        }
        Binding binding = packageBinding.getTypeOrPackage(name);
        if (binding == null) {
            return new ProblemReferenceBinding(CharOperation.arrayConcat(packageBinding.compoundName, name), null, 1);
        }
        if (!binding.isValidBinding()) {
            return new ProblemReferenceBinding((binding instanceof ReferenceBinding) ? ((ReferenceBinding)binding).compoundName : CharOperation.arrayConcat(packageBinding.compoundName, name), (binding instanceof ReferenceBinding) ? ((ReferenceBinding)((ReferenceBinding)binding).closestMatch()) : null, binding.problemId());
        }
        ReferenceBinding typeBinding = (ReferenceBinding)binding;
        if (!typeBinding.canBeSeenBy(this)) {
            return new ProblemReferenceBinding(typeBinding.compoundName, typeBinding, 2);
        }
        return typeBinding;
    }
    
    public final TypeBinding getType(char[][] compoundName, int typeNameLength) {
        if (typeNameLength == 1) {
            TypeBinding binding = getBaseType(compoundName[0]);
            if (binding != null) {
                return binding;
            }
        }
        CompilationUnitScope unitScope = this.compilationUnitScope();
        unitScope.recordQualifiedReference(compoundName);
        Binding binding2 = this.getTypeOrPackage(compoundName[0], (typeNameLength == 1) ? 4 : 20, true);
        if (binding2 == null) {
            char[][] qName2 = { compoundName[0] };
            return new ProblemReferenceBinding(qName2, this.environment().createMissingType(this.compilationUnitScope().getCurrentPackage(), qName2), 1);
        }
        if (!binding2.isValidBinding()) {
            if ((binding2 instanceof PackageBinding)) {
                char[][] qName2 = { compoundName[0] };
                return new ProblemReferenceBinding(qName2, this.environment().createMissingType(null, qName2), 1);
            }
            return (ReferenceBinding)binding2;
        }
        else {
            int currentIndex = 1;
            boolean checkVisibility = false;
            if ((binding2 instanceof PackageBinding)) {
                PackageBinding packageBinding = (PackageBinding)binding2;
                while (currentIndex < typeNameLength) {
                    binding2 = packageBinding.getTypeOrPackage(compoundName[currentIndex++]);
                    if (binding2 == null) {
                        char[][] qName = CharOperation.subarray(compoundName, 0, currentIndex);
                        return new ProblemReferenceBinding(qName, this.environment().createMissingType(packageBinding, qName), 1);
                    }
                    if (!binding2.isValidBinding()) {
                        return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), (binding2 instanceof ReferenceBinding) ? ((ReferenceBinding)((ReferenceBinding)binding2).closestMatch()) : null, binding2.problemId());
                    }
                    if (!(binding2 instanceof PackageBinding)) {
                        break;
                    }
                    packageBinding = (PackageBinding)binding2;
                }
                if ((binding2 instanceof PackageBinding)) {
                    char[][] qName = CharOperation.subarray(compoundName, 0, currentIndex);
                    return new ProblemReferenceBinding(qName, this.environment().createMissingType(null, qName), 1);
                }
                checkVisibility = true;
            }
            ReferenceBinding typeBinding = (ReferenceBinding)binding2;
            unitScope.recordTypeReference(typeBinding);
            if (checkVisibility && !typeBinding.canBeSeenBy(this)) {
                return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), typeBinding, 2);
            }
            while (currentIndex < typeNameLength) {
                typeBinding = this.getMemberType(compoundName[currentIndex++], typeBinding);
                if (!typeBinding.isValidBinding()) {
                    if ((typeBinding instanceof ProblemReferenceBinding)) {
                        ProblemReferenceBinding problemBinding = (ProblemReferenceBinding)typeBinding;
                        return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), problemBinding.closestReferenceMatch(), typeBinding.problemId());
                    }
                    return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), (ReferenceBinding)((ReferenceBinding)binding2).closestMatch(), typeBinding.problemId());
                }
                else {
                    continue;
                }
            }
            return typeBinding;
        }
    }
    
    final Binding getTypeOrPackage(char[] name, int mask, boolean needResolve) {
        Scope scope = this;
        MethodScope methodScope = null;
        ReferenceBinding foundType = null;
        boolean insideStaticContext = false;
        boolean insideTypeAnnotation = false;
        Label_0737: {
            if ((mask & 0x4) == 0x0) {
                for (Scope next = scope; (next = scope.parent) != null; scope = next) {}
            }
            else {
                boolean inheritedHasPrecedence = this.compilerOptions().complianceLevel >= 3145728L;
                while (true) {
                    switch (scope.kind) {
                        case 2: {
                            methodScope = (MethodScope)scope;
                            AbstractMethodDeclaration methodDecl = methodScope.referenceMethod();
                            if (methodDecl != null) {
                                if (methodDecl.binding != null) {
                                    TypeVariableBinding typeVariable = methodDecl.binding.getTypeVariable(name);
                                    if (typeVariable != null) {
                                        return typeVariable;
                                    }
                                }
                                else {
                                    TypeParameter[] params = methodDecl.typeParameters();
                                    int i = (params == null) ? 0 : params.length;
                                    while (--i >= 0) {
                                        if (CharOperation.equals(params[i].name, name) && params[i].binding != null && params[i].binding.isValidBinding()) {
                                            return params[i].binding;
                                        }
                                        continue;
                                    }
                                }
                            }
                            insideStaticContext |= methodScope.isStatic;
                            insideTypeAnnotation = methodScope.insideTypeAnnotation;
                        }
                        case 1: {
                            ReferenceBinding localType = ((BlockScope)scope).findLocalType(name);
                            if (localType != null) {
                                if (foundType != null && TypeBinding.notEquals(foundType, localType)) {
                                    return new ProblemReferenceBinding(new char[][] { name }, foundType, 5);
                                }
                                return localType;
                            }
                            else {
                                break;
                            }
                        }
                        case 3: {
                            SourceTypeBinding sourceType = ((ClassScope)scope).referenceContext.binding;
                            if (scope == this && (sourceType.tagBits & 0x40000L) == 0x0L) {
                                TypeVariableBinding typeVariable2 = sourceType.getTypeVariable(name);
                                if (typeVariable2 != null) {
                                    return typeVariable2;
                                }
                                if (CharOperation.equals(name, sourceType.sourceName)) {
                                    return sourceType;
                                }
                                insideStaticContext |= sourceType.isStatic();
                                break;
                            }
                            else {
                                if (!insideTypeAnnotation) {
                                    ReferenceBinding memberType = this.findMemberType(name, sourceType);
                                    if (memberType != null) {
                                        if (memberType.problemId() == 3) {
                                            if (foundType == null || foundType.problemId() == 2) {
                                                return memberType;
                                            }
                                            return new ProblemReferenceBinding(new char[][] { name }, foundType, 5);
                                        }
                                        else {
                                            if (memberType.isValidBinding() && (TypeBinding.equalsEquals(sourceType, memberType.enclosingType()) || inheritedHasPrecedence)) {
                                                if (insideStaticContext && !memberType.isStatic() && sourceType.isGenericType()) {
                                                    return new ProblemReferenceBinding(new char[][] { name }, memberType, 7);
                                                }
                                                if (foundType == null || (inheritedHasPrecedence && foundType.problemId() == 2)) {
                                                    return memberType;
                                                }
                                                if (foundType.isValidBinding() && TypeBinding.notEquals(foundType, memberType)) {
                                                    return new ProblemReferenceBinding(new char[][] { name }, foundType, 5);
                                                }
                                            }
                                            if (foundType == null || (foundType.problemId() == 2 && memberType.problemId() != 2)) {
                                                foundType = memberType;
                                            }
                                        }
                                    }
                                }
                                TypeVariableBinding typeVariable2 = sourceType.getTypeVariable(name);
                                if (typeVariable2 != null) {
                                    if (insideStaticContext) {
                                        return new ProblemReferenceBinding(new char[][] { name }, typeVariable2, 7);
                                    }
                                    return typeVariable2;
                                }
                                else {
                                    insideStaticContext |= sourceType.isStatic();
                                    insideTypeAnnotation = false;
                                    if (CharOperation.equals(sourceType.sourceName, name)) {
                                        if (foundType != null && TypeBinding.notEquals(foundType, sourceType) && foundType.problemId() != 2) {
                                            return new ProblemReferenceBinding(new char[][] { name }, foundType, 5);
                                        }
                                        return sourceType;
                                    }
                                    else {
                                        break;
                                    }
                                }
                            }
                        }
                        case 4: {
                            if (foundType != null && foundType.problemId() != 2) {
                                return foundType;
                            }
                            break Label_0737;
                        }
                    }
                    scope = scope.parent;
                }
            }
        }
        CompilationUnitScope unitScope = (CompilationUnitScope)scope;
        HashtableOfObject typeOrPackageCache = unitScope.typeOrPackageCache;
        if (typeOrPackageCache != null) {
            Binding cachedBinding = (Binding)typeOrPackageCache.get(name);
            if (cachedBinding != null) {
                if ((cachedBinding instanceof ImportBinding)) {
                    ImportReference importReference = ((ImportBinding)cachedBinding).reference;
                    if (importReference != null) {
                        importReference.bits |= 0x2;
                    }
                    if ((cachedBinding instanceof ImportConflictBinding)) {
                        typeOrPackageCache.put(name, cachedBinding = ((ImportConflictBinding)cachedBinding).conflictingTypeBinding);
                    }
                    else {
                        typeOrPackageCache.put(name, cachedBinding = ((ImportBinding)cachedBinding).resolvedImport);
                    }
                }
                if ((mask & 0x4) != 0x0) {
                    if (foundType != null && foundType.problemId() != 2 && cachedBinding.problemId() != 3) {
                        return foundType;
                    }
                    if ((cachedBinding instanceof ReferenceBinding)) {
                        return cachedBinding;
                    }
                }
                if ((mask & 0x10) != 0x0 && cachedBinding instanceof PackageBinding) {
                    return cachedBinding;
                }
            }
        }
        if ((mask & 0x4) != 0x0) {
            ImportBinding[] imports = unitScope.imports;
            if (imports != null && typeOrPackageCache == null) {
                for (int i = 0, length = imports.length; i < length; ++i) {
                    ImportBinding importBinding = imports[i];
                    if (!importBinding.onDemand && CharOperation.equals(importBinding.compoundName[importBinding.compoundName.length - 1], name)) {
                        Binding resolvedImport = unitScope.resolveSingleImport(importBinding, 4);
                        if (resolvedImport != null && resolvedImport instanceof TypeBinding) {
                            ImportReference importReference2 = importBinding.reference;
                            if (importReference2 != null) {
                                importReference2.bits |= 0x2;
                                return resolvedImport;
                            }
                            return resolvedImport;
                        }
                    }
                }
            }
            PackageBinding currentPackage = unitScope.fPackage;
            unitScope.recordReference(currentPackage.compoundName, name);
            Binding binding = currentPackage.getTypeOrPackage(name);
            if ((binding instanceof ReferenceBinding)) {
                ReferenceBinding referenceType = (ReferenceBinding)binding;
                if ((referenceType.tagBits & 0x80L) == 0x0L) {
                    if (typeOrPackageCache != null) {
                        typeOrPackageCache.put(name, referenceType);
                        return referenceType;
                    }
                    return referenceType;
                }
            }
            if (imports != null) {
                boolean foundInImport = false;
                ReferenceBinding type = null;
                for (int j = 0, length2 = imports.length; j < length2; ++j) {
                    ImportBinding someImport = imports[j];
                    if (someImport.onDemand) {
                        Binding resolvedImport2 = someImport.resolvedImport;
                        ReferenceBinding temp = null;
                        if ((resolvedImport2 instanceof PackageBinding)) {
                            temp = this.findType(name, (PackageBinding)resolvedImport2, currentPackage);
                        }
                        else if (someImport.isStatic()) {
                            temp = this.findMemberType(name, (ReferenceBinding)resolvedImport2);
                            if (temp != null && !temp.isStatic()) {
                                temp = null;
                            }
                        }
                        else {
                            temp = this.findDirectMemberType(name, (ReferenceBinding)resolvedImport2);
                        }
                        if (TypeBinding.notEquals(temp, type) && temp != null) {
                            if (temp.isValidBinding()) {
                                ImportReference importReference3 = someImport.reference;
                                if (importReference3 != null) {
                                    importReference3.bits |= 0x2;
                                }
                                if (foundInImport) {
                                    temp = new ProblemReferenceBinding(new char[][] { name }, type, 3);
                                    if (typeOrPackageCache != null) {
                                        typeOrPackageCache.put(name, temp);
                                        return temp;
                                    }
                                    return temp;
                                }
                                else {
                                    type = temp;
                                    foundInImport = true;
                                }
                            }
                            else if (foundType == null) {
                                foundType = temp;
                            }
                        }
                    }
                }
                if (type != null) {
                    if (typeOrPackageCache != null) {
                        typeOrPackageCache.put(name, type);
                        return type;
                    }
                    return type;
                }
            }
        }
        unitScope.recordSimpleReference(name);
        if ((mask & 0x10) != 0x0) {
            PackageBinding packageBinding = unitScope.environment.getTopLevelPackage(name);
            if (packageBinding != null && (packageBinding.tagBits & 0x80L) == 0x0L) {
                if (typeOrPackageCache != null) {
                    typeOrPackageCache.put(name, packageBinding);
                    return packageBinding;
                }
                return packageBinding;
            }
        }
        if (foundType == null) {
            char[][] qName = { name };
            ReferenceBinding closestMatch = null;
            if ((mask & 0x10) != 0x0) {
                if (needResolve) {
                    closestMatch = this.environment().createMissingType(unitScope.fPackage, qName);
                }
            }
            else {
                PackageBinding packageBinding2 = unitScope.environment.getTopLevelPackage(name);
                if ((packageBinding2 == null || !packageBinding2.isValidBinding()) && needResolve) {
                    closestMatch = this.environment().createMissingType(unitScope.fPackage, qName);
                }
            }
            foundType = new ProblemReferenceBinding(qName, closestMatch, 1);
            if (typeOrPackageCache != null && (mask & 0x10) != 0x0) {
                typeOrPackageCache.put(name, foundType);
            }
        }
        else if ((foundType.tagBits & 0x80L) != 0x0L) {
            char[][] qName = { name };
            foundType = new ProblemReferenceBinding(qName, foundType, 1);
            if (typeOrPackageCache != null && (mask & 0x10) != 0x0) {
                typeOrPackageCache.put(name, foundType);
            }
        }
        return foundType;
    }
    
    public final Binding getTypeOrPackage(char[][] compoundName) {
        int nameLength = compoundName.length;
        if (nameLength == 1) {
            TypeBinding binding = getBaseType(compoundName[0]);
            if (binding != null) {
                return binding;
            }
        }
        Binding binding2 = this.getTypeOrPackage(compoundName[0], 20, true);
        if (!binding2.isValidBinding()) {
            return binding2;
        }
        int currentIndex = 1;
        boolean checkVisibility = false;
        if ((binding2 instanceof PackageBinding)) {
            PackageBinding packageBinding = (PackageBinding)binding2;
            while (currentIndex < nameLength) {
                binding2 = packageBinding.getTypeOrPackage(compoundName[currentIndex++]);
                if (binding2 == null) {
                    return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), null, 1);
                }
                if (!binding2.isValidBinding()) {
                    return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), (binding2 instanceof ReferenceBinding) ? ((ReferenceBinding)((ReferenceBinding)binding2).closestMatch()) : null, binding2.problemId());
                }
                if (!(binding2 instanceof PackageBinding)) {
                    break;
                }
                packageBinding = (PackageBinding)binding2;
            }
            if ((binding2 instanceof PackageBinding)) {
                return binding2;
            }
            checkVisibility = true;
        }
        ReferenceBinding typeBinding = (ReferenceBinding)binding2;
        ReferenceBinding qualifiedType = (ReferenceBinding)this.environment().convertToRawType(typeBinding, false);
        if (checkVisibility && !typeBinding.canBeSeenBy(this)) {
            return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), typeBinding, 2);
        }
        while (currentIndex < nameLength) {
            typeBinding = this.getMemberType(compoundName[currentIndex++], typeBinding);
            if (!typeBinding.isValidBinding()) {
                return new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), (ReferenceBinding)typeBinding.closestMatch(), typeBinding.problemId());
            }
            if (typeBinding.isGenericType()) {
                qualifiedType = this.environment().createRawType(typeBinding, qualifiedType);
            }
            else {
                qualifiedType = ((qualifiedType != null && (qualifiedType.isRawType() || qualifiedType.isParameterizedType())) ? this.environment().createParameterizedType(typeBinding, null, qualifiedType) : typeBinding);
            }
        }
        return qualifiedType;
    }
    
    public boolean hasErasedCandidatesCollisions(TypeBinding one, TypeBinding two, Map invocations, ReferenceBinding type, ASTNode typeRef) {
        invocations.clear();
        TypeBinding[] mecs = this.minimalErasedCandidates(new TypeBinding[] { one, two }, invocations);
        if (mecs != null) {
            for (int k = 0, max = mecs.length; k < max; ++k) {
                TypeBinding mec = mecs[k];
                if (mec != null) {
                    Object value = invocations.get(mec);
                    if ((value instanceof TypeBinding[])) {
                        TypeBinding[] invalidInvocations = (TypeBinding[])value;
                        this.problemReporter().superinterfacesCollide(invalidInvocations[0].erasure(), typeRef, invalidInvocations[0], invalidInvocations[1]);
                        type.tagBits |= 0x20000L;
                        return true;
                    }
                }
            }
        }
        return false;
    }
    
    public CaseStatement innermostSwitchCase() {
        Scope scope = this;
        while (!(scope instanceof BlockScope)) {
            scope = scope.parent;
            if (scope != null) {
                continue;
            }
            return null;
        }
        return ((BlockScope)scope).enclosingCase;
    }
    
    protected boolean isAcceptableMethod(MethodBinding one, MethodBinding two) {
        TypeBinding[] oneParams = one.parameters;
        TypeBinding[] twoParams = two.parameters;
        int oneParamsLength = oneParams.length;
        int twoParamsLength = twoParams.length;
        if (oneParamsLength == twoParamsLength) {
            boolean applyErasure = this.environment().globalOptions.sourceLevel < 3211264L;
            int j = 0;
            while (j < oneParamsLength) {
                TypeBinding oneParam = applyErasure ? oneParams[j].erasure() : oneParams[j];
                TypeBinding twoParam = applyErasure ? twoParams[j].erasure() : twoParams[j];
                if (TypeBinding.equalsEquals(oneParam, twoParam) || oneParam.isCompatibleWith(twoParam)) {
                    if (!two.declaringClass.isRawType()) {
                        TypeBinding leafComponentType = two.original().parameters[j].leafComponentType();
                        TypeBinding originalTwoParam = applyErasure ? leafComponentType.erasure() : leafComponentType;
                        switch (originalTwoParam.kind()) {
                            case 4100: {
                                if (((TypeVariableBinding)originalTwoParam).hasOnlyRawBounds()) {
                                    break;
                                }
                            }
                            case 260:
                            case 516:
                            case 8196: {
                                TypeBinding originalOneParam = one.original().parameters[j].leafComponentType();
                                switch (originalOneParam.kind()) {
                                    case 4:
                                    case 2052: {
                                        TypeBinding inheritedTwoParam = oneParam.findSuperTypeOriginatingFrom(twoParam);
                                        if (inheritedTwoParam != null) {
                                            if (!inheritedTwoParam.leafComponentType().isRawType()) {
                                                break;
                                            }
                                            return false;
                                        }
                                        else {
                                            break;
                                        }
                                    }
                                    case 4100: {
                                        if (!((TypeVariableBinding)originalOneParam).upperBound().isRawType()) {
                                            break;
                                        }
                                        return false;
                                    }
                                    case 1028: {
                                        return false;
                                    }
                                    default: {
                                        break;
                                    }
                                }
                                break;
                            }
                        }
                    }
                    Label_0451: {
                        ++j;
                    }
                }
                else {
                    if (j == oneParamsLength - 1 && one.isVarargs() && two.isVarargs()) {
                        TypeBinding oType = ((ArrayBinding)oneParam).elementsType();
                        TypeBinding eType = ((ArrayBinding)twoParam).elementsType();
                        if (CompilerOptions.tolerateIllegalAmbiguousVarargsInvocation && this.compilerOptions().complianceLevel < 3342336L) {
                            if (TypeBinding.equalsEquals(oneParam, eType) || oneParam.isCompatibleWith(eType)) {
                                return true;
                            }
                        }
                        else if (TypeBinding.equalsEquals(oType, eType) || oType.isCompatibleWith(eType)) {
                            return true;
                        }
                    }
                    return false;
                }
            }
            return true;
        }
        if (one.isVarargs() && two.isVarargs()) {
            if (CompilerOptions.tolerateIllegalAmbiguousVarargsInvocation && this.compilerOptions().complianceLevel < 3342336L && oneParamsLength > twoParamsLength && ((ArrayBinding)twoParams[twoParamsLength - 1]).elementsType().id != 1) {
                return false;
            }
            int i = ((oneParamsLength > twoParamsLength) ? twoParamsLength : oneParamsLength) - 2;
            while (i >= 0) {
                if (TypeBinding.notEquals(oneParams[i], twoParams[i]) && !oneParams[i].isCompatibleWith(twoParams[i])) {
                    return false;
                }
                --i;
            }
            if (this.parameterCompatibilityLevel(one, twoParams, true) == -1 && this.parameterCompatibilityLevel(two, oneParams, true) == 2) {
                return true;
            }
        }
        return false;
    }
    
    public boolean isBoxingCompatibleWith(TypeBinding expressionType, TypeBinding targetType) {
        LookupEnvironment environment = this.environment();
        if (environment.globalOptions.sourceLevel < 3211264L || expressionType.isBaseType() == targetType.isBaseType()) {
            return false;
        }
        TypeBinding convertedType = environment.computeBoxingType(expressionType);
        return TypeBinding.equalsEquals(convertedType, targetType) || convertedType.isCompatibleWith(targetType, this);
    }
    
    public final boolean isDefinedInField(FieldBinding field) {
        Scope scope = this;
        do {
            if ((scope instanceof MethodScope)) {
                MethodScope methodScope = (MethodScope)scope;
                if (methodScope.initializedField == field) {
                    return true;
                }
            }
            scope = scope.parent;
        } while (scope != null);
        return false;
    }
    
    public final boolean isDefinedInMethod(MethodBinding method) {
        method = method.original();
        Scope scope = this;
        do {
            if ((scope instanceof MethodScope)) {
                ReferenceContext refContext = ((MethodScope)scope).referenceContext;
                if (refContext instanceof AbstractMethodDeclaration && ((AbstractMethodDeclaration)refContext).binding == method) {
                    return true;
                }
            }
            scope = scope.parent;
        } while (scope != null);
        return false;
    }
    
    public final boolean isDefinedInSameUnit(ReferenceBinding type) {
        ReferenceBinding enclosingType;
        for (enclosingType = type; (type = enclosingType.enclosingType()) != null; enclosingType = type) {}
        Scope unitScope;
        Scope scope;
        for (unitScope = this; (scope = unitScope.parent) != null; unitScope = scope) {}
        SourceTypeBinding[] topLevelTypes = ((CompilationUnitScope)unitScope).topLevelTypes;
        int i = topLevelTypes.length;
        while (--i >= 0) {
            if (TypeBinding.equalsEquals(topLevelTypes[i], enclosingType.original())) {
                return true;
            }
            continue;
        }
        return false;
    }
    
    public final boolean isDefinedInType(ReferenceBinding type) {
        Scope scope = this;
        while (!(scope instanceof ClassScope) || !TypeBinding.equalsEquals(((ClassScope)scope).referenceContext.binding, type)) {
            scope = scope.parent;
            if (scope != null) {
                continue;
            }
            return false;
        }
        return true;
    }
    
    public boolean isInsideCase(CaseStatement caseStatement) {
        Scope scope = this;
        do {
            switch (scope.kind) {
                case 1: {
                    if (((BlockScope)scope).enclosingCase == caseStatement) {
                        return true;
                    }
                    break;
                }
            }
            scope = scope.parent;
        } while (scope != null);
        return false;
    }
    
    public boolean isInsideDeprecatedCode() {
        switch (this.kind) {
            case 1:
            case 2: {
                MethodScope methodScope = this.methodScope();
                if (!methodScope.isInsideInitializer()) {
                    ReferenceContext referenceContext = methodScope.referenceContext;
                    MethodBinding context = (referenceContext instanceof AbstractMethodDeclaration) ? ((AbstractMethodDeclaration)referenceContext).binding : ((LambdaExpression)referenceContext).binding;
                    if (context != null && context.isViewedAsDeprecated()) {
                        return true;
                    }
                }
                else if (methodScope.initializedField != null && methodScope.initializedField.isViewedAsDeprecated()) {
                    return true;
                }
                SourceTypeBinding declaringType = ((BlockScope)this).referenceType().binding;
                if (declaringType != null) {
                    declaringType.initializeDeprecatedAnnotationTagBits();
                    if (declaringType.isViewedAsDeprecated()) {
                        return true;
                    }
                    break;
                }
                else {
                    break;
                }
            }
            case 3: {
                ReferenceBinding context2 = ((ClassScope)this).referenceType().binding;
                if (context2 != null) {
                    context2.initializeDeprecatedAnnotationTagBits();
                    if (context2.isViewedAsDeprecated()) {
                        return true;
                    }
                    break;
                }
                else {
                    break;
                }
            }
            case 4: {
                CompilationUnitDeclaration unit = this.referenceCompilationUnit();
                if (unit.types != null && unit.types.length > 0) {
                    SourceTypeBinding type = unit.types[0].binding;
                    if (type != null) {
                        type.initializeDeprecatedAnnotationTagBits();
                        if (type.isViewedAsDeprecated()) {
                            return true;
                        }
                        break;
                    }
                    else {
                        break;
                    }
                }
                else {
                    break;
                }
            }
        }
        return false;
    }
    
    private boolean isOverriddenMethodGeneric(MethodBinding method) {
        MethodVerifier verifier = this.environment().methodVerifier();
        for (ReferenceBinding currentType = method.declaringClass.superclass(); currentType != null; currentType = currentType.superclass()) {
            MethodBinding[] currentMethods = currentType.getMethods(method.selector);
            int i = 0;
            int l = currentMethods.length;
            while (i < l) {
                MethodBinding currentMethod = currentMethods[i];
                if (currentMethod != null && currentMethod.original().typeVariables != Binding.NO_TYPE_VARIABLES && verifier.doesMethodOverride(method, currentMethod)) {
                    return true;
                }
                ++i;
            }
        }
        return false;
    }
    
    public boolean isSubtypeOfRawType(TypeBinding paramType) {
        TypeBinding t = paramType.leafComponentType();
        if (t.isBaseType()) {
            return false;
        }
        ReferenceBinding currentType = (ReferenceBinding)t;
        ReferenceBinding[] interfacesToVisit = null;
        int nextPosition = 0;
        while (!currentType.isRawType()) {
            ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
            if (itsInterfaces != null && itsInterfaces != Binding.NO_SUPERINTERFACES) {
                if (interfacesToVisit == null) {
                    interfacesToVisit = itsInterfaces;
                    nextPosition = interfacesToVisit.length;
                }
                else {
                    int itsLength = itsInterfaces.length;
                    if (nextPosition + itsLength >= interfacesToVisit.length) {
                        System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
                    }
                    int a = 0;
                Label_0166:
                    while (a < itsLength) {
                        ReferenceBinding next = itsInterfaces[a];
                        int b = 0;
                        Label_0163: {
                            while (b < nextPosition) {
                                if (TypeBinding.equalsEquals(next, interfacesToVisit[b])) {
                                    ++a;
                                    continue Label_0166;
                                }
                                ++b;
                            }
                            interfacesToVisit[nextPosition++] = next;
                            continue;
                        }
                    }
                }
            }
            if ((currentType = currentType.superclass()) != null) {
                continue;
            }
            int i = 0;
            while (i < nextPosition) {
                currentType = interfacesToVisit[i];
                if (currentType.isRawType()) {
                    return true;
                }
                ReferenceBinding[] itsInterfaces2 = currentType.superInterfaces();
                if (itsInterfaces2 != null && itsInterfaces2 != Binding.NO_SUPERINTERFACES) {
                    int itsLength2 = itsInterfaces2.length;
                    if (nextPosition + itsLength2 >= interfacesToVisit.length) {
                        System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength2 + 5], 0, nextPosition);
                    }
                    int a2 = 0;
                Label_0318:
                    while (a2 < itsLength2) {
                        ReferenceBinding next2 = itsInterfaces2[a2];
                        int b2 = 0;
                        Label_0315: {
                            while (b2 < nextPosition) {
                                if (TypeBinding.equalsEquals(next2, interfacesToVisit[b2])) {
                                    ++a2;
                                    continue Label_0318;
                                }
                                ++b2;
                            }
                            interfacesToVisit[nextPosition++] = next2;
                            continue;
                        }
                    }
                }
                ++i;
            }
            return false;
        }
        return true;
    }
    
    private TypeBinding leastContainingInvocation(TypeBinding mec, Object invocationData, ArrayList lubStack) {
        if (invocationData == null) {
            return mec;
        }
        if ((invocationData instanceof TypeBinding)) {
            return (TypeBinding)invocationData;
        }
        TypeBinding[] invocations = (TypeBinding[])invocationData;
        int dim = mec.dimensions();
        mec = mec.leafComponentType();
        int argLength = mec.typeVariables().length;
        if (argLength == 0) {
            return mec;
        }
        TypeBinding[] bestArguments = new TypeBinding[argLength];
        for (int i = 0, length = invocations.length; i < length; ++i) {
            TypeBinding invocation = invocations[i].leafComponentType();
            switch (invocation.kind()) {
                case 2052: {
                    TypeVariableBinding[] invocationVariables = invocation.typeVariables();
                    int j = 0;
                    while (j < argLength) {
                        TypeBinding bestArgument = this.leastContainingTypeArgument(bestArguments[j], invocationVariables[j], (ReferenceBinding)mec, j, (ArrayList)lubStack.clone());
                        if (bestArgument == null) {
                            return null;
                        }
                        bestArguments[j] = bestArgument;
                        ++j;
                    }
                    break;
                }
                case 260: {
                    ParameterizedTypeBinding parameterizedType = (ParameterizedTypeBinding)invocation;
                    int k = 0;
                    while (k < argLength) {
                        TypeBinding bestArgument2 = this.leastContainingTypeArgument(bestArguments[k], parameterizedType.arguments[k], (ReferenceBinding)mec, k, (ArrayList)lubStack.clone());
                        if (bestArgument2 == null) {
                            return null;
                        }
                        bestArguments[k] = bestArgument2;
                        ++k;
                    }
                    break;
                }
                case 1028: {
                    return (dim == 0) ? invocation : this.environment().createArrayType(invocation, dim);
                }
            }
        }
        TypeBinding least = this.environment().createParameterizedType((ReferenceBinding)mec.erasure(), bestArguments, mec.enclosingType());
        return (dim == 0) ? least : this.environment().createArrayType(least, dim);
    }
    
    private TypeBinding leastContainingTypeArgument(TypeBinding u, TypeBinding v, ReferenceBinding genericType, int rank, ArrayList lubStack) {
        if (u == null) {
            return v;
        }
        if (TypeBinding.equalsEquals(u, v)) {
            return u;
        }
        if (v.isWildcard()) {
            WildcardBinding wildV = (WildcardBinding)v;
            if (u.isWildcard()) {
                WildcardBinding wildU2 = (WildcardBinding)u;
                switch (wildU2.boundKind) {
                    case 1: {
                        switch (wildV.boundKind) {
                            case 1: {
                                TypeBinding lub2 = this.lowerUpperBound(new TypeBinding[] { wildU2.bound, wildV.bound }, lubStack);
                                if (lub2 == null) {
                                    return null;
                                }
                                if (TypeBinding.equalsEquals(lub2, TypeBinding.INT)) {
                                    return this.environment().createWildcard(genericType, rank, null, null, 0);
                                }
                                return this.environment().createWildcard(genericType, rank, lub2, null, 1);
                            }
                            case 2: {
                                if (TypeBinding.equalsEquals(wildU2.bound, wildV.bound)) {
                                    return wildU2.bound;
                                }
                                return this.environment().createWildcard(genericType, rank, null, null, 0);
                            }
                            default: {
                                break;
                            }
                        }
                    }
                    case 2: {
                        if (wildU2.boundKind == 2) {
                            TypeBinding[] glb = greaterLowerBound(new TypeBinding[] { wildU2.bound, wildV.bound }, this, this.environment());
                            if (glb == null) {
                                return null;
                            }
                            return this.environment().createWildcard(genericType, rank, glb[0], null, 2);
                        }
                    }
                    default: {
                        break;
                    }
                }
            }
            else {
                switch (wildV.boundKind) {
                    case 1: {
                        TypeBinding lub = this.lowerUpperBound(new TypeBinding[] { u, wildV.bound }, lubStack);
                        if (lub == null) {
                            return null;
                        }
                        if (TypeBinding.equalsEquals(lub, TypeBinding.INT)) {
                            return this.environment().createWildcard(genericType, rank, null, null, 0);
                        }
                        return this.environment().createWildcard(genericType, rank, lub, null, 1);
                    }
                    case 2: {
                        TypeBinding[] glb = greaterLowerBound(new TypeBinding[] { u, wildV.bound }, this, this.environment());
                        if (glb == null) {
                            return null;
                        }
                        return this.environment().createWildcard(genericType, rank, glb[0], null, 2);
                    }
                    default: {
                        break;
                    }
                }
            }
        }
        else if (u.isWildcard()) {
            WildcardBinding wildU = (WildcardBinding)u;
            switch (wildU.boundKind) {
                case 1: {
                    TypeBinding lub = this.lowerUpperBound(new TypeBinding[] { wildU.bound, v }, lubStack);
                    if (lub == null) {
                        return null;
                    }
                    if (TypeBinding.equalsEquals(lub, TypeBinding.INT)) {
                        return this.environment().createWildcard(genericType, rank, null, null, 0);
                    }
                    return this.environment().createWildcard(genericType, rank, lub, null, 1);
                }
                case 2: {
                    TypeBinding[] glb = greaterLowerBound(new TypeBinding[] { wildU.bound, v }, this, this.environment());
                    if (glb == null) {
                        return null;
                    }
                    return this.environment().createWildcard(genericType, rank, glb[0], null, 2);
                }
                default: {
                    break;
                }
            }
        }
        TypeBinding lub3 = this.lowerUpperBound(new TypeBinding[] { u, v }, lubStack);
        if (lub3 == null) {
            return null;
        }
        if (TypeBinding.equalsEquals(lub3, TypeBinding.INT)) {
            return this.environment().createWildcard(genericType, rank, null, null, 0);
        }
        return this.environment().createWildcard(genericType, rank, lub3, null, 1);
    }
    
    public TypeBinding lowerUpperBound(TypeBinding[] types) {
        int typeLength = types.length;
        if (typeLength == 1) {
            TypeBinding type = types[0];
            return (type == null) ? TypeBinding.VOID : type;
        }
        return this.lowerUpperBound(types, new ArrayList(1));
    }
    
    private TypeBinding lowerUpperBound(TypeBinding[] types, ArrayList lubStack) {
        int typeLength = types.length;
        if (typeLength == 1) {
            TypeBinding type2 = types[0];
            return (type2 == null) ? TypeBinding.VOID : type2;
        }
    Label_0154:
        for (int stackLength = lubStack.size(), i = 0; i < stackLength; ++i) {
            TypeBinding[] lubTypes = (TypeBinding[])lubStack.get(i);
            int lubTypeLength = lubTypes.length;
            if (lubTypeLength >= typeLength) {
            Label_0141:
                for (TypeBinding type : types) {
                    if (type != null) {
                        for (TypeBinding lubType : lubTypes) {
                            if (lubType != null) {
                                if (!TypeBinding.equalsEquals(lubType, type)) {
                                    if (lubType.isEquivalentTo(type)) {
                                        continue Label_0141;
                                    }
                                }
                                else {
                                    continue Label_0141;
                                }
                            }
                        }
                        continue Label_0154;
                    }
                }
                return TypeBinding.INT;
            }
        }
        lubStack.add(types);
        Map<Object, Object> invocations = new HashMap(1);
        TypeBinding[] mecs = this.minimalErasedCandidates(types, invocations);
        if (mecs == null) {
            return null;
        }
        int length = mecs.length;
        if (length == 0) {
            return TypeBinding.VOID;
        }
        int count = 0;
        TypeBinding firstBound = null;
        int commonDim = -1;
        for (TypeBinding mec : mecs) {
            if (mec != null) {
                mec = this.leastContainingInvocation(mec, invocations.get(mec), lubStack);
                if (mec == null) {
                    return null;
                }
                int dim = mec.dimensions();
                if (commonDim == -1) {
                    commonDim = dim;
                }
                else if (dim != commonDim) {
                    return null;
                }
                if (firstBound == null && !mec.leafComponentType().isInterface()) {
                    firstBound = mec.leafComponentType();
                }
                mecs[count++] = mec;
            }
        }
        switch (count) {
            case 0: {
                return TypeBinding.VOID;
            }
            case 1: {
                return mecs[0];
            }
            case 2: {
                if (((commonDim == 0) ? mecs[1].id : mecs[1].leafComponentType().id) == 1) {
                    return mecs[0];
                }
                if (((commonDim == 0) ? mecs[0].id : mecs[0].leafComponentType().id) == 1) {
                    return mecs[1];
                }
                break;
            }
        }
        TypeBinding[] otherBounds = new TypeBinding[count - 1];
        int rank = 0;
        for (int m = 0; m < count; ++m) {
            TypeBinding mec2 = (commonDim == 0) ? mecs[m] : mecs[m].leafComponentType();
            if (mec2.isInterface()) {
                otherBounds[rank++] = mec2;
            }
        }
        TypeBinding intersectionType = this.environment().createWildcard(null, 0, firstBound, otherBounds, 1);
        return (commonDim == 0) ? intersectionType : this.environment().createArrayType(intersectionType, commonDim);
    }
    
    public final MethodScope methodScope() {
        Scope scope = this;
        while (!(scope instanceof MethodScope)) {
            scope = scope.parent;
            if (scope != null) {
                continue;
            }
            return null;
        }
        return (MethodScope)scope;
    }
    
    public final MethodScope namedMethodScope() {
        Scope scope = this;
        while (!(scope instanceof MethodScope) || scope.isLambdaScope()) {
            scope = scope.parent;
            if (scope != null) {
                continue;
            }
            return null;
        }
        return (MethodScope)scope;
    }
    
    protected TypeBinding[] minimalErasedCandidates(TypeBinding[] types, Map allInvocations) {
        int length = types.length;
        int indexOfFirst = -1;
        int actualLength = 0;
        for (int i = 0; i < length; ++i) {
            TypeBinding type = types[i];
            if (type == TypeBinding.NULL) {
                type = (types[i] = null);
            }
            if (type != null) {
                if (type.isBaseType()) {
                    return null;
                }
                if (indexOfFirst < 0) {
                    indexOfFirst = i;
                }
                ++actualLength;
            }
        }
        switch (actualLength) {
            case 0: {
                return Binding.NO_TYPES;
            }
            case 1: {
                return types;
            }
            default: {
                TypeBinding firstType = types[indexOfFirst];
                if (firstType.isBaseType()) {
                    return null;
                }
                ArrayList typesToVisit = new ArrayList(5);
                int dim = firstType.dimensions();
                TypeBinding leafType = firstType.leafComponentType();
                TypeBinding firstErasure;
                switch (leafType.kind()) {
                    case 68:
                    case 260:
                    case 1028: {
                        firstErasure = firstType.erasure();
                        break;
                    }
                    default: {
                        firstErasure = firstType;
                        break;
                    }
                }
                if (TypeBinding.notEquals(firstErasure, firstType)) {
                    allInvocations.put(firstErasure, firstType);
                }
                typesToVisit.add(firstType);
                for (int max = 1, j = 0; j < max; ++j) {
                    TypeBinding typeToVisit = (TypeBinding)typesToVisit.get(j);
                    dim = typeToVisit.dimensions();
                    if (dim > 0) {
                        leafType = typeToVisit.leafComponentType();
                        switch (leafType.id) {
                            case 1: {
                                if (dim > 1) {
                                    TypeBinding elementType = ((ArrayBinding)typeToVisit).elementsType();
                                    if (!typesToVisit.contains(elementType)) {
                                        typesToVisit.add(elementType);
                                        ++max;
                                        continue;
                                    }
                                    continue;
                                }
                            }
                            case 2:
                            case 3:
                            case 4:
                            case 5:
                            case 7:
                            case 8:
                            case 9:
                            case 10: {
                                TypeBinding superType = this.getJavaIoSerializable();
                                if (!typesToVisit.contains(superType)) {
                                    typesToVisit.add(superType);
                                    ++max;
                                }
                                superType = this.getJavaLangCloneable();
                                if (!typesToVisit.contains(superType)) {
                                    typesToVisit.add(superType);
                                    ++max;
                                }
                                superType = this.getJavaLangObject();
                                if (!typesToVisit.contains(superType)) {
                                    typesToVisit.add(superType);
                                    ++max;
                                    continue;
                                }
                                continue;
                            }
                            default: {
                                typeToVisit = leafType;
                                break;
                            }
                        }
                    }
                    ReferenceBinding currentType = (ReferenceBinding)typeToVisit;
                    if (currentType.isCapture()) {
                        TypeBinding firstBound = ((CaptureBinding)currentType).firstBound;
                        if (firstBound != null && firstBound.isArrayType()) {
                            TypeBinding superType2 = (dim == 0) ? firstBound : this.environment().createArrayType(firstBound, dim);
                            if (!typesToVisit.contains(superType2)) {
                                typesToVisit.add(superType2);
                                ++max;
                                TypeBinding superTypeErasure = (firstBound.isTypeVariable() || firstBound.isWildcard()) ? superType2 : superType2.erasure();
                                if (TypeBinding.notEquals(superTypeErasure, superType2)) {
                                    allInvocations.put(superTypeErasure, superType2);
                                    continue;
                                }
                                continue;
                            }
                            else {
                                continue;
                            }
                        }
                    }
                    ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
                    if (itsInterfaces != null) {
                        for (int k = 0, count = itsInterfaces.length; k < count; ++k) {
                            TypeBinding itsInterface = itsInterfaces[k];
                            TypeBinding superType3 = (dim == 0) ? itsInterface : this.environment().createArrayType(itsInterface, dim);
                            if (!typesToVisit.contains(superType3)) {
                                typesToVisit.add(superType3);
                                ++max;
                                TypeBinding superTypeErasure2 = (itsInterface.isTypeVariable() || itsInterface.isWildcard()) ? superType3 : superType3.erasure();
                                if (TypeBinding.notEquals(superTypeErasure2, superType3)) {
                                    allInvocations.put(superTypeErasure2, superType3);
                                }
                            }
                        }
                    }
                    TypeBinding itsSuperclass = currentType.superclass();
                    if (itsSuperclass != null) {
                        TypeBinding superType4 = (dim == 0) ? itsSuperclass : this.environment().createArrayType(itsSuperclass, dim);
                        if (!typesToVisit.contains(superType4)) {
                            typesToVisit.add(superType4);
                            ++max;
                            TypeBinding superTypeErasure3 = (itsSuperclass.isTypeVariable() || itsSuperclass.isWildcard()) ? superType4 : superType4.erasure();
                            if (TypeBinding.notEquals(superTypeErasure3, superType4)) {
                                allInvocations.put(superTypeErasure3, superType4);
                            }
                        }
                    }
                }
                int superLength = typesToVisit.size();
                TypeBinding[] erasedSuperTypes = new TypeBinding[superLength];
                int rank = 0;
                for (TypeBinding type2 : typesToVisit) {
                    leafType = type2.leafComponentType();
                    erasedSuperTypes[rank++] = ((leafType.isTypeVariable() || leafType.isWildcard()) ? type2 : type2.erasure());
                }
                int remaining = superLength;
                for (int l = indexOfFirst + 1; l < length; ++l) {
                    TypeBinding otherType = types[l];
                    if (otherType != null) {
                        if (otherType.isArrayType()) {
                        Label_1187:
                            for (int m = 0; m < superLength; ++m) {
                                TypeBinding erasedSuperType = erasedSuperTypes[m];
                                if (erasedSuperType != null && !TypeBinding.equalsEquals(erasedSuperType, otherType)) {
                                    TypeBinding match;
                                    if ((match = otherType.findSuperTypeOriginatingFrom(erasedSuperType)) == null) {
                                        erasedSuperTypes[m] = null;
                                        if (--remaining == 0) {
                                            return null;
                                        }
                                    }
                                    else {
                                        Object invocationData = allInvocations.get(erasedSuperType);
                                        if (invocationData == null) {
                                            allInvocations.put(erasedSuperType, match);
                                        }
                                        else if ((invocationData instanceof TypeBinding)) {
                                            if (TypeBinding.notEquals(match, (TypeBinding)invocationData)) {
                                                TypeBinding[] someInvocations = { (TypeBinding)invocationData, match };
                                                allInvocations.put(erasedSuperType, someInvocations);
                                            }
                                        }
                                        else {
                                            TypeBinding[] someInvocations = (TypeBinding[])invocationData;
                                            int invocLength = someInvocations.length;
                                            int k2 = 0;
                                            while (k2 < invocLength) {
                                                if (TypeBinding.equalsEquals(someInvocations[k2], match)) {
                                                    continue Label_1187;
                                                }
                                                ++k2;
                                            }
                                            System.arraycopy(someInvocations, 0, someInvocations = new TypeBinding[invocLength + 1], 0, invocLength);
                                            allInvocations.put(erasedSuperType, someInvocations);
                                            someInvocations[invocLength] = match;
                                        }
                                    }
                                }
                            }
                        }
                        else {
                        Label_1463:
                            for (int m = 0; m < superLength; ++m) {
                                TypeBinding erasedSuperType = erasedSuperTypes[m];
                                if (erasedSuperType != null) {
                                    TypeBinding match;
                                    if (TypeBinding.equalsEquals(erasedSuperType, otherType) || (erasedSuperType.id == 1 && otherType.isInterface())) {
                                        match = erasedSuperType;
                                    }
                                    else {
                                        if (erasedSuperType.isArrayType()) {
                                            match = null;
                                        }
                                        else {
                                            match = otherType.findSuperTypeOriginatingFrom(erasedSuperType);
                                        }
                                        if (match == null) {
                                            erasedSuperTypes[m] = null;
                                            if (--remaining == 0) {
                                                return null;
                                            }
                                            continue;
                                        }
                                    }
                                    Object invocationData = allInvocations.get(erasedSuperType);
                                    if (invocationData == null) {
                                        allInvocations.put(erasedSuperType, match);
                                    }
                                    else if ((invocationData instanceof TypeBinding)) {
                                        if (TypeBinding.notEquals(match, (TypeBinding)invocationData)) {
                                            TypeBinding[] someInvocations = { (TypeBinding)invocationData, match };
                                            allInvocations.put(erasedSuperType, someInvocations);
                                        }
                                    }
                                    else {
                                        TypeBinding[] someInvocations = (TypeBinding[])invocationData;
                                        int invocLength = someInvocations.length;
                                        int k2 = 0;
                                        while (k2 < invocLength) {
                                            if (TypeBinding.equalsEquals(someInvocations[k2], match)) {
                                                continue Label_1463;
                                            }
                                            ++k2;
                                        }
                                        System.arraycopy(someInvocations, 0, someInvocations = new TypeBinding[invocLength + 1], 0, invocLength);
                                        allInvocations.put(erasedSuperType, someInvocations);
                                        someInvocations[invocLength] = match;
                                    }
                                }
                            }
                        }
                    }
                }
                if (remaining > 1) {
                    for (int l = 0; l < superLength; ++l) {
                        TypeBinding erasedSuperType2 = erasedSuperTypes[l];
                        if (erasedSuperType2 != null) {
                            for (int m = 0; m < superLength; ++m) {
                                if (l != m) {
                                    TypeBinding otherType2 = erasedSuperTypes[m];
                                    if (otherType2 != null) {
                                        if ((erasedSuperType2 instanceof ReferenceBinding)) {
                                            if ((otherType2.id != 1 || !erasedSuperType2.isInterface()) && erasedSuperType2.findSuperTypeOriginatingFrom(otherType2) != null) {
                                                erasedSuperTypes[m] = null;
                                                --remaining;
                                            }
                                        }
                                        else if (erasedSuperType2.isArrayType() && (!otherType2.isArrayType() || otherType2.leafComponentType().id != 1 || otherType2.dimensions() != erasedSuperType2.dimensions() || !erasedSuperType2.leafComponentType().isInterface()) && erasedSuperType2.findSuperTypeOriginatingFrom(otherType2) != null) {
                                            erasedSuperTypes[m] = null;
                                            --remaining;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    return erasedSuperTypes;
                }
                return erasedSuperTypes;
            }
        }
    }
    
    protected final MethodBinding mostSpecificClassMethodBinding(MethodBinding[] visible, int visibleSize, InvocationSite invocationSite) {
        MethodBinding previous = null;
        int i = 0;
    Label_0112:
        while (i < visibleSize) {
            MethodBinding method = visible[i];
            if (previous != null && TypeBinding.notEquals(method.declaringClass, previous.declaringClass)) {
                break;
            }
            if (!method.isStatic()) {
                previous = method;
            }
            for (int j = 0; j < visibleSize; ++j) {
                if (i != j && !visible[j].areParametersCompatibleWith(method.parameters)) {
                    ++i;
                    continue Label_0112;
                }
            }
            this.compilationUnitScope().recordTypeReferences(method.thrownExceptions);
            return method;
        }
        return new ProblemMethodBinding(visible[0], visible[0].selector, visible[0].parameters, 3);
    }
    
    protected final MethodBinding mostSpecificInterfaceMethodBinding(MethodBinding[] visible, int visibleSize, InvocationSite invocationSite) {
        int i = 0;
    Label_0073:
        while (i < visibleSize) {
            MethodBinding method = visible[i];
            for (int j = 0; j < visibleSize; ++j) {
                if (i != j && !visible[j].areParametersCompatibleWith(method.parameters)) {
                    ++i;
                    continue Label_0073;
                }
            }
            this.compilationUnitScope().recordTypeReferences(method.thrownExceptions);
            return method;
        }
        return new ProblemMethodBinding(visible[0], visible[0].selector, visible[0].parameters, 3);
    }
    
    protected final MethodBinding mostSpecificMethodBinding(MethodBinding[] visible, int visibleSize, TypeBinding[] argumentTypes, final InvocationSite invocationSite, ReferenceBinding receiverType) {
        boolean isJdk18 = this.compilerOptions().sourceLevel >= 3407872L;
        if (isJdk18 && invocationSite.checkingPotentialCompatibility()) {
            if (visibleSize != visible.length) {
                System.arraycopy(visible, 0, visible = new MethodBinding[visibleSize], 0, visibleSize);
            }
            invocationSite.acceptPotentiallyCompatibleMethods(visible);
        }
        int[] compatibilityLevels = new int[visibleSize];
        int compatibleCount = 0;
        for (int i = 0; i < visibleSize; ++i) {
            if ((compatibilityLevels[i] = this.parameterCompatibilityLevel(visible[i], argumentTypes, invocationSite)) != -1) {
                if (i != compatibleCount) {
                    visible[compatibleCount] = visible[i];
                    compatibilityLevels[compatibleCount] = compatibilityLevels[i];
                }
                ++compatibleCount;
            }
        }
        if (compatibleCount == 0) {
            return new ProblemMethodBinding(visible[0].selector, argumentTypes, 1);
        }
        if (compatibleCount == 1) {
            MethodBinding candidate3 = visible[0];
            if (candidate3 != null) {
                this.compilationUnitScope().recordTypeReferences(candidate3.thrownExceptions);
                return candidate3;
            }
            return candidate3;
        }
        else {
            if (compatibleCount != visibleSize) {
                System.arraycopy(visible, 0, visible = new MethodBinding[visibleSize = compatibleCount], 0, compatibleCount);
                System.arraycopy(compatibilityLevels, 0, compatibilityLevels = new int[compatibleCount], 0, compatibleCount);
            }
            MethodBinding[] moreSpecific = new MethodBinding[visibleSize];
            if (isJdk18) {
                int count2 = 0;
                int m = 0;
            Label_0626:
                while (m < visibleSize) {
                    MethodBinding mbj = visible[m].genericMethod();
                    TypeBinding[] mbjParameters = mbj.parameters;
                    int levelj = compatibilityLevels[m];
                    Label_0623: {
                        for (int k2 = 0; k2 < visibleSize; ++k2) {
                            if (m != k2) {
                                int levelk = compatibilityLevels[k2];
                                if (levelj > -1 && levelk > -1 && levelj != levelk) {
                                    if (levelj < levelk) {
                                        continue;
                                    }
                                }
                                else {
                                    MethodBinding mbk = visible[k2].genericMethod();
                                    TypeBinding[] mbkParameters = mbk.parameters;
                                    if ((invocationSite instanceof Invocation || invocationSite instanceof ReferenceExpression) && mbk.typeVariables() != Binding.NO_TYPE_VARIABLES) {
                                        Expression[] expressions = null;
                                        if ((invocationSite instanceof Invocation)) {
                                            expressions = ((Invocation)invocationSite).arguments();
                                        }
                                        else {
                                            expressions = ((ReferenceExpression)invocationSite).createPseudoExpressions(argumentTypes);
                                        }
                                        InferenceContext18 ic18 = new InferenceContext18(this, expressions, null, null);
                                        if (ic18.isMoreSpecificThan(mbj, mbk, levelj == 2, levelk == 2)) {
                                            continue;
                                        }
                                    }
                                    else {
                                        for (int i2 = 0, length = argumentTypes.length; i2 < length; ++i2) {
                                            TypeBinding argumentType = argumentTypes[i2];
                                            TypeBinding s = InferenceContext18.getParameter(mbjParameters, i2, levelj == 2);
                                            TypeBinding t = InferenceContext18.getParameter(mbkParameters, i2, levelk == 2);
                                            if (!TypeBinding.equalsEquals(s, t) && !argumentType.sIsMoreSpecific(s, t, this)) {
                                                break Label_0623;
                                            }
                                        }
                                        if (levelj == 2 && levelk == 2) {
                                            TypeBinding s2 = InferenceContext18.getParameter(mbjParameters, argumentTypes.length, true);
                                            TypeBinding t2 = InferenceContext18.getParameter(mbkParameters, argumentTypes.length, true);
                                            if (!TypeBinding.notEquals(s2, t2) || !t2.isSubtypeOf(s2)) {
                                                continue;
                                            }
                                        }
                                        else {
                                            continue;
                                        }
                                    }
                                }
                                ++m;
                                continue Label_0626;
                            }
                        }
                        moreSpecific[count2++] = visible[m];
                        continue Label_0623;
                    }
                }
                if (count2 == 0) {
                    return new ProblemMethodBinding(visible[0], visible[0].selector, visible[0].parameters, 3);
                }
                if (count2 == 1) {
                    MethodBinding candidate2 = moreSpecific[0];
                    if (candidate2 != null) {
                        this.compilationUnitScope().recordTypeReferences(candidate2.thrownExceptions);
                        return candidate2;
                    }
                    return candidate2;
                }
                else {
                    visibleSize = count2;
                }
            }
            else {
                InvocationSite tieBreakInvocationSite = new InvocationSite(this, invocationSite) {
                    
                    @Override
                    public TypeBinding[] genericTypeArguments() {
                        return null;
                    }
                    
                    @Override
                    public boolean isSuperAccess() {
                        return invocationSite.isSuperAccess();
                    }
                    
                    @Override
                    public boolean isTypeAccess() {
                        return invocationSite.isTypeAccess();
                    }
                    
                    @Override
                    public void setActualReceiverType(ReferenceBinding actualReceiverType) {
                    }
                    
                    @Override
                    public void setDepth(int depth) {
                    }
                    
                    @Override
                    public void setFieldIndex(int depth) {
                    }
                    
                    @Override
                    public int sourceStart() {
                        return invocationSite.sourceStart();
                    }
                    
                    @Override
                    public int sourceEnd() {
                        return invocationSite.sourceStart();
                    }
                    
                    @Override
                    public TypeBinding invocationTargetType() {
                        return invocationSite.invocationTargetType();
                    }
                    
                    @Override
                    public boolean receiverIsImplicitThis() {
                        return invocationSite.receiverIsImplicitThis();
                    }
                    
                    @Override
                    public InferenceContext18 freshInferenceContext(Scope scope) {
                        return null;
                    }
                    
                    @Override
                    public ExpressionContext getExpressionContext() {
                        return ExpressionContext.VANILLA_CONTEXT;
                    }
                    
                    @Override
                    public boolean isQualifiedSuper() {
                        return invocationSite.isQualifiedSuper();
                    }
                    
                    @Override
                    public boolean checkingPotentialCompatibility() {
                        return false;
                    }
                    
                    @Override
                    public void acceptPotentiallyCompatibleMethods(MethodBinding[] methods) {
                    }
                };
                int count = 0;
                for (int level = 0, max = 2; level <= max; ++level) {
                Label_0963:
                    for (int j = 0; j < visibleSize; ++j) {
                        if (compatibilityLevels[j] == level) {
                            max = level;
                            MethodBinding current = visible[j];
                            MethodBinding original = current.original();
                            MethodBinding tiebreakMethod = current.tiebreakMethod();
                            for (int k = 0; k < visibleSize; ++k) {
                                if (j != k && compatibilityLevels[k] == level) {
                                    MethodBinding next = visible[k];
                                    if (original == next.original()) {
                                        compatibilityLevels[k] = -1;
                                    }
                                    else {
                                        MethodBinding methodToTest = next;
                                        if ((next instanceof ParameterizedGenericMethodBinding)) {
                                            ParameterizedGenericMethodBinding pNext = (ParameterizedGenericMethodBinding)next;
                                            if (!pNext.isRaw || pNext.isStatic()) {
                                                methodToTest = pNext.originalMethod;
                                            }
                                        }
                                        MethodBinding acceptable = this.computeCompatibleMethod(methodToTest, tiebreakMethod.parameters, tieBreakInvocationSite, level == 2);
                                        if (acceptable != null) {
                                            if (!acceptable.isValidBinding()) {
                                                continue Label_0963;
                                            }
                                            if (!this.isAcceptableMethod(tiebreakMethod, acceptable)) {
                                                continue Label_0963;
                                            }
                                            if (current.isBridge() && !next.isBridge() && tiebreakMethod.areParametersEqual(acceptable)) {
                                                continue Label_0963;
                                            }
                                        }
                                        else {
                                            continue Label_0963;
                                        }
                                    }
                                }
                            }
                            moreSpecific[j] = current;
                            ++count;
                        }
                    }
                }
                if (count == 1) {
                    int l = 0;
                    while (l < visibleSize) {
                        if (moreSpecific[l] != null) {
                            MethodBinding candidate = visible[l];
                            if (candidate != null) {
                                this.compilationUnitScope().recordTypeReferences(candidate.thrownExceptions);
                                return candidate;
                            }
                            return candidate;
                        }
                        else {
                            ++l;
                        }
                    }
                }
                else if (count == 0) {
                    return new ProblemMethodBinding(visible[0], visible[0].selector, visible[0].parameters, 3);
                }
            }
            if (receiverType != null) {
                receiverType = ((receiverType instanceof CaptureBinding) ? receiverType : ((ReferenceBinding)receiverType.erasure()));
            }
        Label_1867:
            for (int i3 = 0; i3 < visibleSize; ++i3) {
                MethodBinding current2 = moreSpecific[i3];
                if (current2 != null) {
                    ReferenceBinding[] mostSpecificExceptions = null;
                    MethodBinding original2 = current2.original();
                    boolean shouldIntersectExceptions = original2.declaringClass.isAbstract() && original2.thrownExceptions != Binding.NO_EXCEPTIONS;
                    for (int j2 = 0; j2 < visibleSize; ++j2) {
                        MethodBinding next2 = moreSpecific[j2];
                        if (next2 != null && i3 != j2) {
                            MethodBinding original3 = next2.original();
                            if (TypeBinding.equalsEquals(original2.declaringClass, original3.declaringClass)) {
                                break Label_1867;
                            }
                            if (!original2.isAbstract()) {
                                if (!original3.isAbstract() && !original3.isDefaultMethod()) {
                                    original3 = original2.findOriginalInheritedMethod(original3);
                                    if (original3 == null) {
                                        continue Label_1867;
                                    }
                                    if ((current2.hasSubstitutedParameters() || original2.typeVariables != Binding.NO_TYPE_VARIABLES) && !this.environment().methodVerifier().isParameterSubsignature(original2, original3)) {
                                        continue Label_1867;
                                    }
                                }
                            }
                            else if (receiverType != null) {
                                TypeBinding superType = receiverType.findSuperTypeOriginatingFrom(original2.declaringClass.erasure());
                                if (!TypeBinding.equalsEquals(original2.declaringClass, superType) && superType instanceof ReferenceBinding) {
                                    MethodBinding[] superMethods = ((ReferenceBinding)superType).getMethods(original2.selector, argumentTypes.length);
                                    int m2 = 0;
                                    int l2 = superMethods.length;
                                    while (m2 < l2) {
                                        if (superMethods[m2].original() == original2) {
                                            original2 = superMethods[m2];
                                            break;
                                        }
                                        ++m2;
                                    }
                                }
                                superType = receiverType.findSuperTypeOriginatingFrom(original3.declaringClass.erasure());
                                if (!TypeBinding.equalsEquals(original3.declaringClass, superType) && superType instanceof ReferenceBinding) {
                                    MethodBinding[] superMethods = ((ReferenceBinding)superType).getMethods(original3.selector, argumentTypes.length);
                                    int m2 = 0;
                                    int l2 = superMethods.length;
                                    while (m2 < l2) {
                                        if (superMethods[m2].original() == original3) {
                                            original3 = superMethods[m2];
                                            break;
                                        }
                                        ++m2;
                                    }
                                }
                                if (original2.typeVariables != Binding.NO_TYPE_VARIABLES) {
                                    original3 = original2.computeSubstitutedMethod(original3, this.environment());
                                }
                                if (original3 != null) {
                                    if (!original2.areParameterErasuresEqual(original3)) {
                                        continue Label_1867;
                                    }
                                    if (TypeBinding.notEquals(original2.returnType, original3.returnType) && !((next2.original().typeVariables == Binding.NO_TYPE_VARIABLES) ? current2.returnType.isCompatibleWith(next2.returnType) : (original2.returnType.erasure().findSuperTypeOriginatingFrom(original3.returnType.erasure()) != null))) {
                                        continue Label_1867;
                                    }
                                    if (shouldIntersectExceptions && original3.declaringClass.isInterface() && current2.thrownExceptions != next2.thrownExceptions) {
                                        if (next2.thrownExceptions == Binding.NO_EXCEPTIONS) {
                                            mostSpecificExceptions = Binding.NO_EXCEPTIONS;
                                        }
                                        else {
                                            if (mostSpecificExceptions == null) {
                                                mostSpecificExceptions = current2.thrownExceptions;
                                            }
                                            int mostSpecificLength = mostSpecificExceptions.length;
                                            ReferenceBinding[] nextExceptions = this.getFilteredExceptions(next2);
                                            int nextLength = nextExceptions.length;
                                            SimpleSet temp = new SimpleSet(mostSpecificLength);
                                            boolean changed = false;
                                            for (ReferenceBinding exception : mostSpecificExceptions) {
                                                int s3 = 0;
                                                while (s3 < nextLength) {
                                                    ReferenceBinding nextException = nextExceptions[s3];
                                                    if (exception.isCompatibleWith(nextException)) {
                                                        temp.add(exception);
                                                        break;
                                                    }
                                                    if (nextException.isCompatibleWith(exception)) {
                                                        temp.add(nextException);
                                                        changed = true;
                                                        break;
                                                    }
                                                    changed = true;
                                                    ++s3;
                                                }
                                            }
                                            if (changed) {
                                                mostSpecificExceptions = ((temp.elementSize == 0) ? Binding.NO_EXCEPTIONS : new ReferenceBinding[temp.elementSize]);
                                                temp.asArray(mostSpecificExceptions);
                                            }
                                        }
                                    }
                                }
                                else {
                                    continue Label_1867;
                                }
                            }
                        }
                    }
                    if (mostSpecificExceptions != null && mostSpecificExceptions != current2.thrownExceptions) {
                        return new MostSpecificExceptionMethodBinding(current2, mostSpecificExceptions);
                    }
                    return current2;
                }
            }
            return new ProblemMethodBinding(visible[0], visible[0].selector, visible[0].parameters, 3);
        }
    }
    
    private ReferenceBinding[] getFilteredExceptions(MethodBinding method) {
        ReferenceBinding[] allExceptions = method.thrownExceptions;
        int length = allExceptions.length;
        if (length < 2) {
            return allExceptions;
        }
        ReferenceBinding[] filteredExceptions = new ReferenceBinding[length];
        int count = 0;
        int i = 0;
    Label_0111:
        while (i < length) {
            ReferenceBinding currentException = allExceptions[i];
            Label_0108: {
                for (int j = 0; j < length; ++j) {
                    if (i != j) {
                        if (TypeBinding.equalsEquals(currentException, allExceptions[j])) {
                            if (i < j) {
                                break;
                            }
                        }
                        else if (!currentException.isCompatibleWith(allExceptions[j])) {
                            continue;
                        }
                        ++i;
                        continue Label_0111;
                    }
                }
                filteredExceptions[count++] = currentException;
                continue;
            }
        }
        if (count != length) {
            ReferenceBinding[] tmp = new ReferenceBinding[count];
            System.arraycopy(filteredExceptions, 0, tmp, 0, count);
            return tmp;
        }
        return allExceptions;
    }
    
    public final ClassScope outerMostClassScope() {
        ClassScope lastClassScope = null;
        Scope scope = this;
        do {
            if ((scope instanceof ClassScope)) {
                lastClassScope = (ClassScope)scope;
            }
            scope = scope.parent;
        } while (scope != null);
        return lastClassScope;
    }
    
    public final MethodScope outerMostMethodScope() {
        MethodScope lastMethodScope = null;
        Scope scope = this;
        do {
            if ((scope instanceof MethodScope)) {
                lastMethodScope = (MethodScope)scope;
            }
            scope = scope.parent;
        } while (scope != null);
        return lastMethodScope;
    }
    
    public int parameterCompatibilityLevel(MethodBinding method, TypeBinding[] arguments, InvocationSite site) {
        if (method.problemId() == 23) {
            method = ((ProblemMethodBinding)method).closestMatch;
            if (method == null) {
                return -1;
            }
        }
        if (this.compilerOptions().sourceLevel >= 3407872L && method instanceof ParameterizedGenericMethodBinding) {
            int inferenceKind = 0;
            InferenceContext18 context = null;
            if ((site instanceof Invocation)) {
                Invocation invocation = (Invocation)site;
                context = invocation.getInferenceContext((ParameterizedGenericMethodBinding)method);
                if (context != null) {
                    inferenceKind = context.inferenceKind;
                }
            }
            else if ((site instanceof ReferenceExpression)) {
                ReferenceExpression referenceExpression = (ReferenceExpression)site;
                context = referenceExpression.getInferenceContext((ParameterizedGenericMethodBinding)method);
                if (context != null) {
                    inferenceKind = context.inferenceKind;
                }
            }
            if (site instanceof Invocation && context != null && context.stepCompleted >= 2) {
                for (int i = 0, length = arguments.length; i < length; ++i) {
                    TypeBinding argument = arguments[i];
                    if (argument.isFunctionalType()) {
                        TypeBinding parameter = InferenceContext18.getParameter(method.parameters, i, context.isVarArgs());
                        if (!argument.isCompatibleWith(parameter, this)) {
                            if (argument.isPolyType()) {
                                parameter = InferenceContext18.getParameter(method.original().parameters, i, context.isVarArgs());
                                if (!((PolyTypeBinding)argument).expression.isPertinentToApplicability(parameter, method)) {
                                    continue;
                                }
                            }
                            return -1;
                        }
                    }
                }
            }
            switch (inferenceKind) {
                case 1: {
                    return 0;
                }
                case 2: {
                    return 1;
                }
                case 3: {
                    return 2;
                }
                default: {
                    break;
                }
            }
        }
        return this.parameterCompatibilityLevel(method, arguments, false);
    }
    
    public int parameterCompatibilityLevel(MethodBinding method, TypeBinding[] arguments) {
        return this.parameterCompatibilityLevel(method, arguments, false);
    }
    
    public int parameterCompatibilityLevel(MethodBinding method, TypeBinding[] arguments, boolean tiebreakingVarargsMethods) {
        TypeBinding[] parameters = method.parameters;
        int paramLength = parameters.length;
        int argLength = arguments.length;
        CompilerOptions compilerOptions = this.compilerOptions();
        if (compilerOptions.sourceLevel < 3211264L) {
            if (paramLength != argLength) {
                return -1;
            }
            int k = 0;
            while (k < argLength) {
                TypeBinding param3 = parameters[k];
                TypeBinding arg3 = arguments[k];
                if (TypeBinding.notEquals(arg3, param3) && !arg3.isCompatibleWith(param3.erasure(), this)) {
                    return -1;
                }
                ++k;
            }
            return 0;
        }
        else {
            if (tiebreakingVarargsMethods && CompilerOptions.tolerateIllegalAmbiguousVarargsInvocation && compilerOptions.complianceLevel < 3342336L) {
                tiebreakingVarargsMethods = false;
            }
            int level = 0;
            int lastIndex = argLength;
            LookupEnvironment env = this.environment();
            if (method.isVarargs()) {
                lastIndex = paramLength - 1;
                if (paramLength == argLength) {
                    TypeBinding param = parameters[lastIndex];
                    TypeBinding arg2 = arguments[lastIndex];
                    if (TypeBinding.notEquals(param, arg2)) {
                        level = this.parameterCompatibilityLevel(arg2, param, env, tiebreakingVarargsMethods);
                        if (level == -1) {
                            param = ((ArrayBinding)param).elementsType();
                            if (tiebreakingVarargsMethods) {
                                arg2 = ((ArrayBinding)arg2).elementsType();
                            }
                            if (this.parameterCompatibilityLevel(arg2, param, env, tiebreakingVarargsMethods) == -1) {
                                return -1;
                            }
                            level = 2;
                        }
                    }
                }
                else {
                    if (paramLength < argLength) {
                        TypeBinding param = ((ArrayBinding)parameters[lastIndex]).elementsType();
                        int i = lastIndex;
                        while (i < argLength) {
                            TypeBinding arg = (tiebreakingVarargsMethods && i == argLength - 1) ? ((ArrayBinding)arguments[i]).elementsType() : arguments[i];
                            if (TypeBinding.notEquals(param, arg) && this.parameterCompatibilityLevel(arg, param, env, tiebreakingVarargsMethods) == -1) {
                                return -1;
                            }
                            ++i;
                        }
                    }
                    else if (lastIndex != argLength) {
                        return -1;
                    }
                    level = 2;
                }
            }
            else if (paramLength != argLength) {
                return -1;
            }
            for (int j = 0; j < lastIndex; ++j) {
                TypeBinding param2 = parameters[j];
                TypeBinding arg = (tiebreakingVarargsMethods && j == argLength - 1) ? ((ArrayBinding)arguments[j]).elementsType() : arguments[j];
                if (TypeBinding.notEquals(arg, param2)) {
                    int newLevel = this.parameterCompatibilityLevel(arg, param2, env, tiebreakingVarargsMethods);
                    if (newLevel == -1) {
                        return -1;
                    }
                    if (newLevel > level) {
                        level = newLevel;
                    }
                }
            }
            return level;
        }
    }
    
    public int parameterCompatibilityLevel(TypeBinding arg, TypeBinding param) {
        if (TypeBinding.equalsEquals(arg, param)) {
            return 0;
        }
        if (arg == null || param == null) {
            return -1;
        }
        if (arg.isCompatibleWith(param, this)) {
            return 0;
        }
        if (arg.kind() == 65540 || arg.isBaseType() != param.isBaseType()) {
            TypeBinding convertedType = this.environment().computeBoxingType(arg);
            if (TypeBinding.equalsEquals(convertedType, param) || convertedType.isCompatibleWith(param, this)) {
                return 1;
            }
        }
        return -1;
    }
    
    private int parameterCompatibilityLevel(TypeBinding arg, TypeBinding param, LookupEnvironment env, boolean tieBreakingVarargsMethods) {
        if (arg == null || param == null) {
            return -1;
        }
        if (arg.isCompatibleWith(param, this)) {
            return 0;
        }
        if (tieBreakingVarargsMethods && (this.compilerOptions().complianceLevel >= 3342336L || !CompilerOptions.tolerateIllegalAmbiguousVarargsInvocation)) {
            return -1;
        }
        if (arg.kind() == 65540 || arg.isBaseType() != param.isBaseType()) {
            TypeBinding convertedType = env.computeBoxingType(arg);
            if (TypeBinding.equalsEquals(convertedType, param) || convertedType.isCompatibleWith(param, this)) {
                return 1;
            }
        }
        return -1;
    }
    
    public abstract ProblemReporter problemReporter();
    
    public final CompilationUnitDeclaration referenceCompilationUnit() {
        Scope unitScope;
        Scope scope;
        for (unitScope = this; (scope = unitScope.parent) != null; unitScope = scope) {}
        return ((CompilationUnitScope)unitScope).referenceContext;
    }
    
    public ReferenceContext referenceContext() {
        Scope current = this;
        do {
            switch (current.kind) {
                case 2: {
                    return ((MethodScope)current).referenceContext;
                }
                case 3: {
                    return ((ClassScope)current).referenceContext;
                }
                case 4: {
                    return ((CompilationUnitScope)current).referenceContext;
                }
                default: {
                    continue;
                }
            }
        } while ((current = current.parent) != null);
        return null;
    }
    
    public ReferenceContext originalReferenceContext() {
        Scope current = this;
        do {
            switch (current.kind) {
                case 2: {
                    ReferenceContext context = ((MethodScope)current).referenceContext;
                    if ((context instanceof LambdaExpression)) {
                        LambdaExpression expression;
                        for (expression = (LambdaExpression)context; expression != expression.original; expression = expression.original) {}
                        return expression;
                    }
                    return context;
                }
                case 3: {
                    return ((ClassScope)current).referenceContext;
                }
                case 4: {
                    return ((CompilationUnitScope)current).referenceContext;
                }
                default: {
                    continue;
                }
            }
        } while ((current = current.parent) != null);
        return null;
    }
    
    public boolean deferCheck(Runnable check) {
        return this.parent != null && this.parent.deferCheck(check);
    }
    
    public void deferBoundCheck(TypeReference typeRef) {
        if (this.kind == 3) {
            ClassScope classScope = (ClassScope)this;
            if (classScope.deferredBoundChecks == null) {
                (classScope.deferredBoundChecks = new ArrayList(3)).add(typeRef);
            }
            else if (!classScope.deferredBoundChecks.contains(typeRef)) {
                classScope.deferredBoundChecks.add(typeRef);
            }
        }
    }
    
    int startIndex() {
        return 0;
    }
    
    public MethodBinding getStaticFactory(ParameterizedTypeBinding allocationType, ReferenceBinding originalEnclosingType, TypeBinding[] argumentTypes, InvocationSite allocationSite) {
        int classTypeVariablesArity = 0;
        TypeVariableBinding[] classTypeVariables = Binding.NO_TYPE_VARIABLES;
        ReferenceBinding currentType;
        ReferenceBinding genericType = currentType = allocationType.genericType();
        while (currentType != null) {
            TypeVariableBinding[] typeVariables = currentType.typeVariables();
            int length = (typeVariables == null) ? 0 : typeVariables.length;
            if (length > 0) {
                System.arraycopy(classTypeVariables, 0, classTypeVariables = new TypeVariableBinding[classTypeVariablesArity + length], 0, classTypeVariablesArity);
                System.arraycopy(typeVariables, 0, classTypeVariables, classTypeVariablesArity, length);
                classTypeVariablesArity += length;
            }
            if (currentType.isStatic()) {
                break;
            }
            currentType = currentType.enclosingType();
        }
        MethodBinding[] methods = allocationType.getMethods(TypeConstants.INIT, argumentTypes.length);
        MethodBinding[] staticFactories = new MethodBinding[methods.length];
        int sfi = 0;
        for (int i = 0, length2 = methods.length; i < length2; ++i) {
            MethodBinding method = methods[i];
            if (method.canBeSeenBy(allocationSite, this)) {
                int paramLength = method.parameters.length;
                boolean isVarArgs = method.isVarargs();
                if (argumentTypes.length != paramLength) {
                    if (isVarArgs) {
                        if (argumentTypes.length < paramLength - 1) {
                            continue;
                        }
                    }
                    else {
                        continue;
                    }
                }
                TypeVariableBinding[] methodTypeVariables = method.typeVariables();
                int methodTypeVariablesArity = methodTypeVariables.length;
                int factoryArity = classTypeVariablesArity + methodTypeVariablesArity;
                LookupEnvironment environment = this.environment();
                MethodBinding staticFactory = new SyntheticFactoryMethodBinding(method.original(), environment, originalEnclosingType);
                staticFactory.typeVariables = new TypeVariableBinding[factoryArity];
                final SimpleLookupTable map = new SimpleLookupTable(factoryArity);
                String prime = "";
                Binding declaringElement = null;
                for (int j = 0; j < classTypeVariablesArity; ++j) {
                    TypeVariableBinding original = classTypeVariables[j];
                    if (original.declaringElement != declaringElement) {
                        declaringElement = original.declaringElement;
                        prime = prime + "'";
                    }
                    map.put(original.unannotated(), staticFactory.typeVariables[j] = new TypeVariableBinding(CharOperation.concat(original.sourceName, prime.toCharArray()), staticFactory, j, environment));
                }
                prime = prime + "'";
                for (int j = classTypeVariablesArity, k = 0; j < factoryArity; ++j, ++k) {
                    map.put(methodTypeVariables[k].unannotated(), staticFactory.typeVariables[j] = new TypeVariableBinding(CharOperation.concat(methodTypeVariables[k].sourceName, prime.toCharArray()), staticFactory, j, environment));
                }
                final Scope scope = this;
                Substitution substitution = new Substitution(this, scope, map) {
                    
                    @Override
                    public LookupEnvironment environment() {
                        return scope.environment();
                    }
                    
                    @Override
                    public boolean isRawSubstitution() {
                        return false;
                    }
                    
                    @Override
                    public TypeBinding substitute(TypeVariableBinding typeVariable) {
                        TypeBinding retVal = (TypeBinding)map.get(typeVariable.unannotated());
                        return (retVal == null) ? typeVariable : (typeVariable.hasTypeAnnotations() ? this.environment().createAnnotatedType(retVal, typeVariable.getTypeAnnotations()) : retVal);
                    }
                };
                for (int l = 0; l < factoryArity; ++l) {
                    TypeVariableBinding originalVariable = (l < classTypeVariablesArity) ? classTypeVariables[l] : methodTypeVariables[l - classTypeVariablesArity];
                    TypeVariableBinding substitutedVariable = (TypeVariableBinding)map.get(originalVariable.unannotated());
                    TypeBinding substitutedSuperclass = substitute(substitution, originalVariable.superclass);
                    ReferenceBinding[] substitutedInterfaces = substitute(substitution, originalVariable.superInterfaces);
                    if (originalVariable.firstBound != null) {
                        TypeBinding firstBound = TypeBinding.equalsEquals(originalVariable.firstBound, originalVariable.superclass) ? substitutedSuperclass : substitutedInterfaces[0];
                        substitutedVariable.setFirstBound(firstBound);
                    }
                    switch (substitutedSuperclass.kind()) {
                        case 68: {
                            substitutedVariable.setSuperClass(environment.getResolvedType(TypeConstants.JAVA_LANG_OBJECT, null));
                            substitutedVariable.setSuperInterfaces(substitutedInterfaces);
                            break;
                        }
                        default: {
                            if (substitutedSuperclass.isInterface()) {
                                substitutedVariable.setSuperClass(environment.getResolvedType(TypeConstants.JAVA_LANG_OBJECT, null));
                                int interfaceCount = substitutedInterfaces.length;
                                System.arraycopy(substitutedInterfaces, 0, substitutedInterfaces = new ReferenceBinding[interfaceCount + 1], 1, interfaceCount);
                                substitutedInterfaces[0] = (ReferenceBinding)substitutedSuperclass;
                                substitutedVariable.setSuperInterfaces(substitutedInterfaces);
                                break;
                            }
                            substitutedVariable.setSuperClass((ReferenceBinding)substitutedSuperclass);
                            substitutedVariable.setSuperInterfaces(substitutedInterfaces);
                            break;
                        }
                    }
                }
                staticFactory.returnType = environment.createParameterizedType(genericType, substitute(substitution, genericType.typeVariables()), originalEnclosingType);
                staticFactory.parameters = substitute(substitution, method.parameters);
                staticFactory.thrownExceptions = substitute(substitution, method.thrownExceptions);
                if (staticFactory.thrownExceptions == null) {
                    staticFactory.thrownExceptions = Binding.NO_EXCEPTIONS;
                }
                staticFactories[sfi++] = new ParameterizedMethodBinding((ParameterizedTypeBinding)environment.convertToParameterizedType(staticFactory.declaringClass), staticFactory);
            }
        }
        if (sfi == 0) {
            return null;
        }
        if (sfi != methods.length) {
            System.arraycopy(staticFactories, 0, staticFactories = new MethodBinding[sfi], 0, sfi);
        }
        MethodBinding[] compatible = new MethodBinding[sfi];
        int compatibleIndex = 0;
        for (int m = 0; m < sfi; ++m) {
            MethodBinding compatibleMethod = this.computeCompatibleMethod(staticFactories[m], argumentTypes, allocationSite);
            if (compatibleMethod != null && compatibleMethod.isValidBinding()) {
                compatible[compatibleIndex++] = compatibleMethod;
            }
        }
        if (compatibleIndex == 0) {
            return null;
        }
        return (compatibleIndex == 1) ? compatible[0] : this.mostSpecificMethodBinding(compatible, compatibleIndex, argumentTypes, allocationSite, allocationType);
    }
    
    public boolean validateNullAnnotation(long tagBits, TypeReference typeRef, Annotation[] annotations) {
        if (typeRef == null) {
            return true;
        }
        TypeBinding type = typeRef.resolvedType;
        boolean usesNullTypeAnnotations = this.environment().usesNullTypeAnnotations();
        long nullAnnotationTagBit;
        if (usesNullTypeAnnotations) {
            type = type.leafComponentType();
            nullAnnotationTagBit = (type.tagBits & 0x180000000000000L);
        }
        else {
            nullAnnotationTagBit = (tagBits & 0x180000000000000L);
        }
        if (nullAnnotationTagBit != 0L && type != null && type.isBaseType()) {
            if (typeRef.resolvedType.id != 6 || !usesNullTypeAnnotations) {
                this.problemReporter().illegalAnnotationForBaseType(typeRef, annotations, nullAnnotationTagBit);
            }
            return false;
        }
        else {
            return true;
        }
    }
    
    public abstract boolean hasDefaultNullnessFor(int p0);
    
    public static BlockScope typeAnnotationsResolutionScope(Scope scope) {
        BlockScope resolutionScope = null;
        switch (scope.kind) {
            case 3: {
                resolutionScope = ((ClassScope)scope).referenceContext.staticInitializerScope;
                break;
            }
            case 1:
            case 2: {
                resolutionScope = (BlockScope)scope;
                break;
            }
        }
        return resolutionScope;
    }
    
    public void tagAsAccessingEnclosingInstanceStateOf(ReferenceBinding enclosingType, boolean typeVariableAccess) {
        MethodScope methodScope = this.methodScope();
        if (methodScope != null && methodScope.referenceContext instanceof TypeDeclaration && !methodScope.enclosingReceiverType().isCompatibleWith(enclosingType)) {
            methodScope = methodScope.enclosingMethodScope();
        }
        MethodBinding enclosingMethod = (enclosingType != null) ? enclosingType.enclosingMethod() : null;
        while (methodScope != null) {
            while (methodScope != null && methodScope.referenceContext instanceof LambdaExpression) {
                LambdaExpression lambda = (LambdaExpression)methodScope.referenceContext;
                if (!typeVariableAccess) {
                    lambda.shouldCaptureInstance = true;
                }
                methodScope = methodScope.enclosingMethodScope();
            }
            if (methodScope != null) {
                if ((methodScope.referenceContext instanceof MethodDeclaration)) {
                    MethodDeclaration methodDeclaration = (MethodDeclaration)methodScope.referenceContext;
                    if (methodDeclaration.binding == enclosingMethod) {
                        break;
                    }
                    methodDeclaration.bits &= 0xFFFFFEFF;
                }
                ClassScope enclosingClassScope = methodScope.enclosingClassScope();
                if (enclosingClassScope != null) {
                    TypeDeclaration type = enclosingClassScope.referenceContext;
                    if (type != null && type.binding != null && enclosingType != null && !type.binding.isCompatibleWith(enclosingType.original())) {
                        methodScope = enclosingClassScope.enclosingMethodScope();
                    }
                    else {
                        break;
                    }
                }
                else {
                    break;
                }
            }
            else {
                continue;
            }
        }
    }
    
    public static class Substitutor
    {
        public ReferenceBinding[] substitute(Substitution substitution, ReferenceBinding[] originalTypes) {
            if (originalTypes == null) {
                return null;
            }
            ReferenceBinding[] substitutedTypes = originalTypes;
            int i = 0;
            int length = originalTypes.length;
            while (i < length) {
                ReferenceBinding originalType = originalTypes[i];
                TypeBinding substitutedType = this.substitute(substitution, originalType);
                if (!(substitutedType instanceof ReferenceBinding)) {
                    return null;
                }
                if (substitutedType != originalType) {
                    if (substitutedTypes == originalTypes) {
                        System.arraycopy(originalTypes, 0, substitutedTypes = new ReferenceBinding[length], 0, i);
                    }
                    substitutedTypes[i] = (ReferenceBinding)substitutedType;
                }
                else if (substitutedTypes != originalTypes) {
                    substitutedTypes[i] = originalType;
                }
                ++i;
            }
            return substitutedTypes;
        }
        
        public TypeBinding substitute(Substitution substitution, TypeBinding originalType) {
            if (originalType == null) {
                return null;
            }
            switch (originalType.kind()) {
                case 4100: {
                    return substitution.substitute((TypeVariableBinding)originalType);
                }
                case 260: {
                    ParameterizedTypeBinding originalParameterizedType = (ParameterizedTypeBinding)originalType;
                    ReferenceBinding substitutedEnclosing;
                    ReferenceBinding originalEnclosing = substitutedEnclosing = originalType.enclosingType();
                    if (originalEnclosing != null) {
                        substitutedEnclosing = (ReferenceBinding)this.substitute(substitution, originalEnclosing);
                        if (isMemberTypeOfRaw(originalType, substitutedEnclosing)) {
                            return originalParameterizedType.environment.createRawType(originalParameterizedType.genericType(), substitutedEnclosing, originalType.getTypeAnnotations());
                        }
                    }
                    TypeBinding[] originalArguments = originalParameterizedType.arguments;
                    TypeBinding[] substitutedArguments;
                    if ((substitutedArguments = originalArguments) != null) {
                        if (substitution.isRawSubstitution()) {
                            return originalParameterizedType.environment.createRawType(originalParameterizedType.genericType(), substitutedEnclosing, originalType.getTypeAnnotations());
                        }
                        substitutedArguments = this.substitute(substitution, originalArguments);
                    }
                    if (substitutedArguments != originalArguments || substitutedEnclosing != originalEnclosing) {
                        return originalParameterizedType.environment.createParameterizedType(originalParameterizedType.genericType(), substitutedArguments, substitutedEnclosing, originalType.getTypeAnnotations());
                    }
                    break;
                }
                case 68: {
                    ArrayBinding originalArrayType = (ArrayBinding)originalType;
                    TypeBinding originalLeafComponentType = originalArrayType.leafComponentType;
                    TypeBinding substitute = this.substitute(substitution, originalLeafComponentType);
                    if (substitute != originalLeafComponentType) {
                        return originalArrayType.environment.createArrayType(substitute.leafComponentType(), substitute.dimensions() + originalType.dimensions(), originalType.getTypeAnnotations());
                    }
                    break;
                }
                case 516:
                case 8196: {
                    WildcardBinding wildcard = (WildcardBinding)originalType;
                    if (wildcard.boundKind != 0) {
                        TypeBinding originalBound = wildcard.bound;
                        TypeBinding substitutedBound = this.substitute(substitution, originalBound);
                        TypeBinding[] originalOtherBounds = wildcard.otherBounds;
                        TypeBinding[] substitutedOtherBounds = this.substitute(substitution, originalOtherBounds);
                        if (substitutedBound != originalBound || originalOtherBounds != substitutedOtherBounds) {
                            if (originalOtherBounds != null) {
                                TypeBinding[] bounds = new TypeBinding[1 + substitutedOtherBounds.length];
                                bounds[0] = substitutedBound;
                                System.arraycopy(substitutedOtherBounds, 0, bounds, 1, substitutedOtherBounds.length);
                                TypeBinding[] glb = Scope.greaterLowerBound(bounds, null, substitution.environment());
                                if (glb != null && glb != bounds) {
                                    substitutedBound = glb[0];
                                    if (glb.length == 1) {
                                        substitutedOtherBounds = null;
                                    }
                                    else {
                                        System.arraycopy(glb, 1, substitutedOtherBounds = new TypeBinding[glb.length - 1], 0, glb.length - 1);
                                    }
                                }
                            }
                            return wildcard.environment.createWildcard(wildcard.genericType, wildcard.rank, substitutedBound, substitutedOtherBounds, wildcard.boundKind, wildcard.getTypeAnnotations());
                        }
                        else {
                            break;
                        }
                    }
                    else {
                        break;
                    }
                }
                case 32772: {
                    IntersectionTypeBinding18 intersection = (IntersectionTypeBinding18)originalType;
                    ReferenceBinding[] types = intersection.getIntersectingTypes();
                    TypeBinding[] substitutes = this.substitute(substitution, types);
                    ReferenceBinding[] refSubsts = new ReferenceBinding[substitutes.length];
                    System.arraycopy(substitutes, 0, refSubsts, 0, substitutes.length);
                    return substitution.environment().createIntersectionType18(refSubsts);
                }
                case 4: {
                    if (!originalType.isMemberType()) {
                        break;
                    }
                    ReferenceBinding originalReferenceType = (ReferenceBinding)originalType.unannotated();
                    ReferenceBinding originalEnclosing = originalType.enclosingType();
                    ReferenceBinding substitutedEnclosing;
                    if ((substitutedEnclosing = originalEnclosing) != null) {
                        substitutedEnclosing = (ReferenceBinding)this.substitute(substitution, originalEnclosing);
                        if (isMemberTypeOfRaw(originalType, substitutedEnclosing)) {
                            return substitution.environment().createRawType(originalReferenceType, substitutedEnclosing, originalType.getTypeAnnotations());
                        }
                    }
                    if (substitutedEnclosing != originalEnclosing) {
                        return substitution.isRawSubstitution() ? substitution.environment().createRawType(originalReferenceType, substitutedEnclosing, originalType.getTypeAnnotations()) : substitution.environment().createParameterizedType(originalReferenceType, null, substitutedEnclosing, originalType.getTypeAnnotations());
                    }
                    break;
                }
                case 2052: {
                    ReferenceBinding originalReferenceType = (ReferenceBinding)originalType.unannotated();
                    ReferenceBinding substitutedEnclosing;
                    ReferenceBinding originalEnclosing = substitutedEnclosing = originalType.enclosingType();
                    if (originalEnclosing != null) {
                        substitutedEnclosing = (ReferenceBinding)(originalType.isStatic() ? substitution.environment().convertToRawType(originalEnclosing, true) : ((TypeBinding)this.substitute(substitution, originalEnclosing)));
                        if (isMemberTypeOfRaw(originalType, substitutedEnclosing)) {
                            return substitution.environment().createRawType(originalReferenceType, substitutedEnclosing, originalType.getTypeAnnotations());
                        }
                    }
                    if (substitution.isRawSubstitution()) {
                        return substitution.environment().createRawType(originalReferenceType, substitutedEnclosing, originalType.getTypeAnnotations());
                    }
                    TypeBinding[] originalArguments = originalReferenceType.typeVariables();
                    TypeBinding[] substitutedArguments = this.substitute(substitution, originalArguments);
                    return substitution.environment().createParameterizedType(originalReferenceType, substitutedArguments, substitutedEnclosing, originalType.getTypeAnnotations());
                }
            }
            return originalType;
        }
        
        private static boolean isMemberTypeOfRaw(TypeBinding originalType, ReferenceBinding substitutedEnclosing) {
            return substitutedEnclosing != null && substitutedEnclosing.isRawType() && originalType instanceof ReferenceBinding && !((ReferenceBinding)originalType).isStatic();
        }
        
        public TypeBinding[] substitute(Substitution substitution, TypeBinding[] originalTypes) {
            if (originalTypes == null) {
                return null;
            }
            TypeBinding[] substitutedTypes = originalTypes;
            for (int i = 0, length = originalTypes.length; i < length; ++i) {
                TypeBinding originalType = originalTypes[i];
                TypeBinding substitutedParameter = this.substitute(substitution, originalType);
                if (substitutedParameter != originalType) {
                    if (substitutedTypes == originalTypes) {
                        System.arraycopy(originalTypes, 0, substitutedTypes = new TypeBinding[length], 0, i);
                    }
                    substitutedTypes[i] = substitutedParameter;
                }
                else if (substitutedTypes != originalTypes) {
                    substitutedTypes[i] = originalType;
                }
            }
            return substitutedTypes;
        }
    }
    
    class MethodClashException extends RuntimeException
    {
        private static final long serialVersionUID = -7996779527641476028L;
    }
}
